# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from hyperspy.signals import Signal1D
import numpy as np

from pyxem.components import ReducedIntensityCorrectionComponent
from scipy import special


class ReducedIntensity1D(Signal1D):
    """Signal class for Reduced Intensity profiles as a function of scattering vector."""

    _signal_type = "reduced_intensity"

    def damp_exponential(self, b, inplace=True, *args, **kwargs):
        """Damps the reduced intensity signal to reduce noise in the high s
        region by a factor of exp(-b*(s^2)), where b is the damping parameter.

        Parameters
        ----------
        b : float
            The damping parameter.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().
        """

        s_scale = self.axes_manager.signal_axes[0].scale
        s_size = self.axes_manager.signal_axes[0].size
        s_offset = self.axes_manager.signal_axes[0].offset

        return self.map(
            _damp_ri_exponential,
            b=b,
            s_scale=s_scale,
            s_size=s_size,
            s_offset=s_offset,
            inplace=inplace,
            *args,
            **kwargs
        )

    def damp_lorch(self, s_max=None, inplace=True, *args, **kwargs):
        """Damps the reduced intensity signal to reduce noise in the high s
        region by a factor of sin(s*delta) / (s*delta), where
        delta = pi / s_max. See [1].

        Parameters
        ----------
        s_max : float
            The maximum s value to be used for transformation to PDF.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().

        References
        ----------
        [1] Lorch, E. (1969). Neutron diffraction by germania, silica and
        radiation-damaged silica glasses. Journal of Physics C: Solid State
        Physics, 2(2), 229.
        """
        s_scale = self.axes_manager.signal_axes[0].scale
        s_size = self.axes_manager.signal_axes[0].size
        s_offset = self.axes_manager.signal_axes[0].offset
        if not s_max:
            s_max = s_scale * s_size + s_offset

        return self.map(
            _damp_ri_lorch,
            s_max=s_max,
            s_scale=s_scale,
            s_size=s_size,
            s_offset=s_offset,
            inplace=inplace,
            *args,
            **kwargs
        )

    def damp_updated_lorch(self, s_max=None, inplace=True, *args, **kwargs):
        """Damps the reduced intensity signal to reduce noise in the high s
        region by a factor of 3 / (s*delta)^3 (sin(s*delta)-s*delta(cos(s*delta))),
        where delta = pi / s_max. From [1].

        Parameters
        ----------
        s_max : float
            the damping parameter, which need not be the maximum scattering
            vector s to be used for the PDF transform. It's a good guess however
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().

        References
        ----------
        [1] Soper, A. K., & Barney, E. R. (2011). Extracting the pair
        distribution function from white-beam X-ray total scattering data.
        Journal of Applied Crystallography, 44(4), 714-726.
        """
        s_scale = self.axes_manager.signal_axes[0].scale
        s_size = self.axes_manager.signal_axes[0].size
        s_offset = self.axes_manager.signal_axes[0].offset
        if not s_max:
            s_max = s_scale * s_size + s_offset

        return self.map(
            _damp_ri_updated_lorch,
            s_max=s_max,
            s_scale=s_scale,
            s_size=s_size,
            s_offset=s_offset,
            inplace=inplace,
            *args,
            **kwargs
        )

    def damp_extrapolate_to_zero(self, s_min, *args, **kwargs):
        """Extrapolates the reduced intensity to zero linearly below s_min.
        This method is always inplace.

        Parameters
        ----------
        s_min : float
            Value of s below which extrapolation to zero is done.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().
        """
        s_scale = self.axes_manager.signal_axes[0].scale
        s_size = self.axes_manager.signal_axes[0].size
        s_offset = self.axes_manager.signal_axes[0].offset

        return self.map(
            _damp_ri_extrapolate_to_zero,
            s_min=s_min,
            s_scale=s_scale,
            s_size=s_size,
            s_offset=s_offset,
            *args,
            **kwargs
        )

    def damp_low_q_region_erfc(
        self, scale=20, offset=1.3, inplace=True, *args, **kwargs
    ):
        """Damps the reduced intensity signal in the low q region as a
        correction to central beam effects. The reduced intensity profile is
        damped by (erf(scale * s - offset) + 1) / 2

        Parameters
        ----------
        scale : float
            A scalar multiplier for s in the error function
        offset : float
            A scalar offset affecting the error function.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().
        """
        s_scale = self.axes_manager.signal_axes[0].scale
        s_size = self.axes_manager.signal_axes[0].size
        s_offset = self.axes_manager.signal_axes[0].offset

        return self.map(
            _damp_ri_low_q_region_erfc,
            scale=scale,
            offset=offset,
            s_scale=s_scale,
            s_size=s_size,
            s_offset=s_offset,
            inplace=inplace,
            *args,
            **kwargs
        )

    def fit_thermal_multiple_scattering_correction(self, s_max=None, plot=False):
        """Fits a 4th order polynomial function to the reduced intensity.
        This is used to calculate the error in the reduced intensity due to
        the effects of multiple and thermal diffuse scattering, which
        results in the earlier background fit being incorrect for either
        low or high angle scattering (or both). A correction is then applied,
        making the reduced intensity oscillate around zero as it should. This
        will distort peak shape. For more detail see [1].

        To use this correction, the fitted data should be fitted to high
        scattering vector, so that the intensity goes to zero at q_max
        (to prevent FFT artifacts).

        Parameters
        ----------
        s_max : float
            Maximum range of fit. The reduced intensity should go to zero
            at this value.
        plot : bool
            Whether to plot the fit after fitting. If True, fit is plotted.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().

        References
        ----------
        [1] Mu, X. et al. (2013). Evolution of order in amorphous-to-crystalline
        phase transformation of MgF2. Journal of Applied Crystallography, 46(4),
        1105-1116.
        """
        s_scale = self.axes_manager.signal_axes[0].scale
        s_size = self.axes_manager.signal_axes[0].size
        s_offset = self.axes_manager.signal_axes[0].offset
        if not s_max:
            s_max = s_scale * (s_size + 1) + s_offset

        # scattering_axis = s_scale * np.arange(s_size,dtype='float64')
        fit_model = self.create_model()
        fit_model.append(ReducedIntensityCorrectionComponent())
        fit_model.set_signal_range([0, s_max])
        fit_model.multifit()
        fit_value = fit_model.as_signal()
        if plot:
            fit_model.plot()

        self.data = self.data - fit_value

        return None


def _damp_ri_exponential(z, b, s_scale, s_size, s_offset, *args, **kwargs):
    """Used by hs.map in the ReducedIntensity1D to damp the reduced
    intensity signal to reduce noise in the high s region by a factor of
    exp(-b*(s^2)), where b is the damping parameter.

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    b : float
        The damping parameter.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    scattering_axis = s_scale * np.arange(s_size, dtype="float64") + s_offset
    damping_term = np.exp(-b * np.square(scattering_axis))
    return z * damping_term


def _damp_ri_lorch(z, s_max, s_scale, s_size, s_offset, *args, **kwargs):
    """Damp the reduced intensity signal to reduce noise in the high s region by a factor of
    sin(s*delta) / (s*delta), where delta = pi / s_max. (from Lorch 1969).

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    s_max : float
        The maximum s value to be used for transformation to PDF.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    delta = np.pi / s_max

    scattering_axis = s_scale * np.arange(s_size, dtype="float64") + s_offset
    damping_term = np.sin(delta * scattering_axis) / (delta * scattering_axis)
    damping_term = np.nan_to_num(damping_term)
    return z * damping_term


def _damp_ri_updated_lorch(z, s_max, s_scale, s_size, s_offset, *args, **kwargs):
    """Damp the reduced intensity signal to reduce noise in the high s region by a factor of
    3 / (s*delta)^3 (sin(s*delta)-s*delta(cos(s*delta))),
    where delta = pi / s_max.

    From "Extracting the pair distribution function from white-beam X-ray
    total scattering data", Soper & Barney, (2011).

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    s_max : float
        The damping parameter, which need not be the maximum scattering
        vector s to be used for the PDF transform.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    delta = np.pi / s_max

    scattering_axis = s_scale * np.arange(s_size, dtype="float64") + s_offset
    exponent_array = 3 * np.ones(scattering_axis.shape)
    cubic_array = np.power(scattering_axis, exponent_array)
    multiplicative_term = np.divide(3 / (delta**3), cubic_array)
    sine_term = np.sin(delta * scattering_axis) - delta * scattering_axis * np.cos(
        delta * scattering_axis
    )

    damping_term = multiplicative_term * sine_term
    damping_term = np.nan_to_num(damping_term)
    return z * damping_term


def _damp_ri_extrapolate_to_zero(z, s_min, s_scale, s_size, s_offset, *args, **kwargs):
    """Extrapolate the reduced intensity signal to zero below s_min.

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    s_min : float
        Value of s below which data is extrapolated to zero.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    s_min_num = int((s_min - s_offset) / s_scale)

    s_min_val = z[s_min_num]
    extrapolated_vals = np.arange(s_min_num) * s_scale + s_offset
    extrapolated_vals *= s_min_val / extrapolated_vals[-1]  # scale zero to one

    z[:s_min_num] = extrapolated_vals

    return z


def _damp_ri_low_q_region_erfc(
    z, scale, offset, s_scale, s_size, s_offset, *args, **kwargs
):
    """Damp the reduced intensity signal in the low q region as a correction to central beam
    effects. The reduced intensity profile is damped by
    (erf(scale * s - offset) + 1) / 2

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    scale : float
        A scalar multiplier for s in the error function
    offset : float
        A scalar offset affecting the error function.
    scale : float
        The scattering vector calibration of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    scattering_axis = s_scale * np.arange(s_size, dtype="float64") + s_offset

    damping_term = (special.erf(scattering_axis * scale - offset) + 1) / 2
    return z * damping_term



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Variance generators in real and reciprocal space for fluctuation electron microscopy."""

import numpy as np
from hyperspy.signals import Signal2D
from hyperspy.api import stack

from pyxem.signals import DiffractionVariance2D, ImageVariance
from pyxem.utils._signals import (
    _transfer_navigation_axes_to_signal_axes,
    _transfer_signal_axes,
)
from pyxem.utils._deprecated import deprecated


class VarianceGenerator:
    """Generates variance images for a specified signal and set of aperture
    positions.

    Parameters
    ----------
    signal : ElectronDiffraction2D
        The signal of electron diffraction patterns to be indexed.

    """

    @deprecated(
        since="0.15",
        alternative="pyxem.signals.diffraction2d.get_variance",
        removal="1.0.0",
    )
    def __init__(self, signal, *args, **kwargs):
        self.signal = signal
        self.thickness_filter = None

        # add a check for calibration

    @deprecated(
        since="0.15",
        alternative="pyxem.signals.diffraction2d.get_variance",
        removal="1.0.0",
    )
    def get_diffraction_variance(self, dqe, set_data_type=None):
        """Calculates the variance in scattered intensity as a function of
        scattering vector.

        Parameters
        ----------
        dqe : float
            Detective quantum efficiency of the detector for Poisson noise
            correction.
        data_type : numpy data type.
            For numpy data types, see
            https://docs.scipy.org/doc/numpy-1.13.0/user/basics.types.html.
            This is incorporated as squaring the numbers in meansq_dp results
            in considerably larger than the ones in the original array. This can
            result in an overflow error that is difficult to distinguish. Hence
            the data can be converted to a different data type to accommodate.

        Returns
        -------
        vardps : DiffractionVariance2D
            A DiffractionVariance2D object containing the mean DP, mean
            squared DP, and variance DP.
        """

        dp = self.signal
        mean_dp = dp.mean((0, 1))
        if set_data_type is None:
            meansq_dp = Signal2D(np.square(dp.data)).mean((0, 1))
        else:
            meansq_dp = Signal2D(np.square(dp.data.astype(set_data_type))).mean((0, 1))

        normvar = (meansq_dp.data / np.square(mean_dp.data)) - 1.0
        var_dp = Signal2D(normvar)
        corr_var_array = var_dp.data - (np.divide(dqe, mean_dp.data))
        corr_var_array[np.isinf(corr_var_array)] = 0
        corr_var_array[np.isnan(corr_var_array)] = 0
        corr_var = Signal2D(corr_var_array)
        vardps = stack((mean_dp, meansq_dp, var_dp, corr_var))
        sig_x = vardps.data.shape[1]
        sig_y = vardps.data.shape[2]

        dv = DiffractionVariance2D(vardps.data.reshape((2, 2, sig_x, sig_y)))

        dv = _transfer_signal_axes(dv, self.signal)

        return dv

    @deprecated(
        since="0.15",
        alternative="pyxem.signals.diffraction2d.get_variance",
        removal="1.0.0",
    )
    def get_image_variance(self, dqe):
        """Calculates the variance in scattered intensity as a function of
        scattering vector. The calculated variance is normalised by the mean
        squared, as is appropriate for the distribution of intensities. This
        causes a problem if Poisson noise is significant in the data, resulting
        in a divergence of the Poisson noise term. To in turn remove this
        effect, we subtract a dqe/mean_dp term (although it is suggested that
        dqe=1) from the data, creating a "poisson noise-free" corrected variance
        pattern. DQE is fitted to make this pattern flat.

        Parameters
        ----------
        dqe : float
            Detective quantum efficiency of the detector for Poisson noise
            correction.

        Returns
        -------
        varims : ImageVariance
            A two dimensional Signal class object containing the mean DP, mean
            squared DP, and variance DP, and a Poisson noise-corrected variance
            DP.
        """
        im = self.signal.T
        mean_im = im.mean((0, 1))
        meansq_im = Signal2D(np.square(im.data)).mean((0, 1))
        normvar = (meansq_im.data / np.square(mean_im.data)) - 1.0
        var_im = Signal2D(normvar)
        corr_var_array = normvar - (np.divide(dqe, mean_im.data))
        corr_var_array[np.invert(np.isfinite(corr_var_array))] = 0
        corr_var = Signal2D(corr_var_array)
        varims = stack((mean_im, meansq_im, var_im, corr_var))

        sig_x = varims.data.shape[1]
        sig_y = varims.data.shape[2]
        iv = ImageVariance(varims.data.reshape((2, 2, sig_x, sig_y)))
        iv = _transfer_navigation_axes_to_signal_axes(iv, self.signal)

        return iv



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.
import warnings

import numpy as np

from hyperspy.signals import Signal2D
from hyperspy._signals.lazy import LazySignal

from pyxem.signals import DiffractionVectors, VDFSegment
from pyxem.utils._signals import _transfer_signal_axes
from pyxem.utils.segment_utils import separate_watershed


class VirtualDarkFieldImage(Signal2D):
    """Signal class for virtual diffraction contrast images."""

    _signal_type = "virtual_dark_field"

    def __init__(self, *args, **kwargs):
        _vectors = kwargs.pop("vectors", None)
        super().__init__(*args, **kwargs)
        self.metadata.add_node("Vectors")
        if _vectors is not None or "Vectors" not in self.metadata:
            self.vectors = kwargs.pop("vectors", None)

    @property
    def vectors(self):
        return self.metadata.Vectors

    @vectors.setter
    def vectors(self, value):
        self.metadata.Vectors = value

    def get_vdf_segments(
        self,
        min_distance=1,
        min_size=1,
        max_size=np.inf,
        max_number_of_grains=np.inf,
        marker_radius=1,
        threshold=False,
        exclude_border=False,
    ):
        """Separate segments from each of the virtual dark field (VDF) images
        using edge-detection by the Sobel transform and the watershed
        segmentation method implemented in scikit-image [1,2]. Obtain a
        VDFSegment, similar to VDFImage, but where each image is a
        segment of a VDF and the vectors correspond to each segment and
        are not necessarily unique.

        Parameters
        ----------
        min_distance: int
            Minimum distance (in pixels) between grains required for
            them to be considered as separate grains.
        min_size : float
            Grains with size (i.e. total number of pixels) below
            min_size are discarded.
        max_size : float
            Grains with size (i.e. total number of pixels) above
            max_size are discarded.
        max_number_of_grains : int
            Maximum number of grains included in the returned separated
            grains. If it is exceeded, those with highest peak
            intensities will be returned.
        marker_radius : float
            If 1 or larger, each marker for watershed is expanded to a disk
            of radius marker_radius. marker_radius should not exceed
            2*min_distance.
        threshold: bool
            If True, a mask is calculated by thresholding the VDF image
            by the Li threshold method in scikit-image. If False
            (default), the mask is the boolean VDF image.
        exclude_border : int or True, optional
            If non-zero integer, peaks within a distance of
            exclude_border from the boarder will be discarded. If True,
            peaks at or closer than min_distance of the boarder, will be
            discarded.

        References
        ----------
        [1] http://scikit-image.org/docs/dev/auto_examples/segmentation/plot_watershed.html
        [2] https://scikit-image.org/docs/dev/auto_examples/segmentation/plot_expand_labels.html#sphx-glr-auto-examples-segmentation-plot-expand-labels-py

        Returns
        -------
        vdfsegs : VDFSegment
            VDFSegment object containing segments (i.e. grains) of
            single virtual dark field images with corresponding vectors.
        """
        warnings.warn(
            "Changed in version 0.15.0.  May cause unexpected "
            "errors related to managing the proper axes."
        )
        vdfs = self.copy()
        vectors = self.vectors.data

        # TODO : Add aperture radius as an attribute of VDFImage?

        # Create an array of length equal to the number of vectors where each
        # element is a object with shape (n: number of segments for this
        # VDFImage, VDFImage size x, VDFImage size y).
        vdfsegs = np.array(
            vdfs.map(
                separate_watershed,
                show_progressbar=True,
                inplace=False,
                min_distance=min_distance,
                min_size=min_size,
                max_size=max_size,
                max_number_of_grains=max_number_of_grains,
                marker_radius=marker_radius,
                threshold=threshold,
                exclude_border=exclude_border,
                ragged=True,
            ),
            dtype=object,
        )

        segments, vectors_of_segments = [], []
        for i, vector in zip(np.arange(vectors.size), vectors):
            segments = np.append(segments, vdfsegs[i])
            num_segs = np.shape(vdfsegs[i])[0]
            vectors_of_segments = np.append(
                vectors_of_segments, np.broadcast_to(vector, (num_segs, 2))
            )

        vectors_of_segments = vectors_of_segments.reshape((-1, 2))
        segments = segments.reshape(
            (
                np.shape(vectors_of_segments)[0],
                vdfs.axes_manager.signal_shape[0],
                vdfs.axes_manager.signal_shape[1],
            )
        )
        # Calculate the total intensities of each segment
        segment_intensities = np.array(
            [[np.sum(x, axis=(0, 1))] for x in segments], dtype="object"
        )

        # if TraitError is raised, it is likely no segments were found
        segments = Signal2D(segments).transpose(navigation_axes=[0], signal_axes=[2, 1])
        # Create VDFSegment and transfer axes calibrations
        vdfsegs = VDFSegment(
            segments, DiffractionVectors(vectors_of_segments), segment_intensities
        )
        vdfsegs.segments = _transfer_signal_axes(vdfsegs.segments, vdfs)
        n = vdfsegs.segments.axes_manager.navigation_axes[0]
        n.name = "n"
        n.units = "number"
        return vdfsegs


class LazyVirtualDarkFieldImage(LazySignal, VirtualDarkFieldImage):
    pass



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from pyxem.utils import polar_transform_utils as ptu
import pytest
import numpy as np
from unittest.mock import Mock

try:
    import cupy as cp

    CUPY_INSTALLED = True
except ImportError:
    CUPY_INSTALLED = False
    cp = np


skip_cupy = pytest.mark.skipif(not CUPY_INSTALLED, reason="cupy is required")


@pytest.fixture()
def mock_simulation():
    mock_sim = Mock()
    mock_sim.calibrated_coordinates = np.array(
        [[3, 4, 0], [5, 12, 0], [8, 15, 0], [7, 24, 0]]  # 5  # 13  # 17
    )  # 25
    mock_sim.intensities = np.array([1, 2, 3, 4])
    return mock_sim


@skip_cupy
@pytest.fixture()
def mock_simulation_gpu():
    mock_sim = Mock()
    mock_sim.calibrated_coordinates = cp.array(
        [[3, 4, 0], [5, 12, 0], [8, 15, 0], [7, 24, 0]]  # 5  # 13  # 17
    )  # 25
    mock_sim.intensities = cp.array([1, 2, 3, 4])
    return mock_sim


@pytest.mark.parametrize(
    "max_r, expected_r",
    [
        (None, np.array([5, 13, 17, 25])),
        (15, np.array([5, 13])),
    ],
)
def test_template_to_polar(mock_simulation, max_r, expected_r):
    r, theta, _ = ptu.get_template_polar_coordinates(mock_simulation, max_r=max_r)
    np.testing.assert_array_almost_equal(r, expected_r)


@skip_cupy
@pytest.mark.parametrize(
    "max_r, expected_r",
    [
        (None, cp.array([5, 13, 17, 25])),
        (15, cp.array([5, 13])),
    ],
)
def test_template_to_polar_gpu(mock_simulation_gpu, max_r, expected_r):
    r, theta, _ = ptu.get_template_polar_coordinates(mock_simulation_gpu, max_r=max_r)
    cp.testing.assert_allclose(r, expected_r)


@pytest.mark.parametrize(
    "angle, window, expected_x, expected_y",
    [
        (0, None, np.array([3, 5, 8, 7]), np.array([4, 12, 15, 24])),
        (90, None, -np.array([4, 12, 15, 24]), np.array([3, 5, 8, 7])),
        (90, (7, 7), np.array([]), np.array([])),
    ],
)
def test_get_template_cartesian_coordinates(
    mock_simulation, angle, window, expected_x, expected_y
):
    x, y, _ = ptu.get_template_cartesian_coordinates(
        mock_simulation, in_plane_angle=angle, window_size=window
    )
    np.testing.assert_array_almost_equal(x, expected_x)
    np.testing.assert_array_almost_equal(y, expected_y)


@skip_cupy
@pytest.mark.parametrize(
    "angle, window, expected_x, expected_y",
    [
        (0, None, cp.array([3, 5, 8, 7]), cp.array([4, 12, 15, 24])),
        (90, None, -cp.array([4, 12, 15, 24]), cp.array([3, 5, 8, 7])),
        (90, (7, 7), cp.array([]), cp.array([])),
    ],
)
def test_get_template_cartesian_coordinates_gpu(
    mock_simulation_gpu, angle, window, expected_x, expected_y
):
    x, y, _ = ptu.get_template_cartesian_coordinates(
        mock_simulation_gpu, in_plane_angle=angle, window_size=window
    )
    cp.testing.assert_array_almost_equal(x, expected_x)
    cp.testing.assert_array_almost_equal(y, expected_y)


@pytest.mark.parametrize(
    "image_shape, max_r, expected",
    [
        ((20, 20), None, (360, 15)),
        ((20, 20), 10, (360, 10)),
    ],
)
def test_get_polar_pattern_shape(image_shape, max_r, expected):
    result = ptu.get_polar_pattern_shape(image_shape, 1, 1, max_r=max_r)
    np.testing.assert_array_almost_equal(result, expected)


@pytest.mark.parametrize(
    "delta_r, delta_theta, max_r, fdb, db, expected_shape",
    [
        (1, 1, None, False, None, (360, 126)),
        (3, 2, None, False, (110, 80), (180, 42)),
        (1, 1, 200, True, None, (360, 200)),
    ],
)
def test_image_to_polar(delta_r, delta_theta, max_r, fdb, db, expected_shape):
    image = np.ones((200, 151))
    result = ptu.image_to_polar(
        image,
        delta_r=delta_r,
        delta_theta=delta_theta,
        max_r=max_r,
        find_direct_beam=fdb,
        direct_beam_position=db,
    )
    np.testing.assert_array_almost_equal(result.shape, expected_shape)


@skip_cupy
@pytest.mark.parametrize(
    "delta_r, delta_theta, max_r, fdb, db, expected_shape",
    [
        (1, 1, None, False, None, (360, 126)),
        (3, 2, None, False, (110, 80), (180, 42)),
        (1, 1, 200, True, None, (360, 200)),
    ],
)
def test_image_to_polar_gpu(delta_r, delta_theta, max_r, fdb, db, expected_shape):
    image = cp.ones((200, 151))
    result = ptu.image_to_polar(
        image,
        delta_r=delta_r,
        delta_theta=delta_theta,
        max_r=max_r,
        find_direct_beam=fdb,
        direct_beam_position=db,
    )
    cp.testing.assert_array_almost_equal(result.shape, expected_shape)


@pytest.mark.parametrize(
    "center, radius, output_shape, expected_shape",
    [
        ((1, 1), 1, (360, 7), (3, 2, 360, 7)),
        ((5, 8), 10, (10, 10), (3, 2, 10, 10)),
    ],
)
def test_chunk_polar(center, radius, output_shape, expected_shape):
    chunk = np.ones((3, 2, 10, 7))
    polar_chunk = ptu._chunk_to_polar(chunk, center, radius, output_shape)
    np.testing.assert_array_almost_equal(polar_chunk.shape, expected_shape)


@skip_cupy
@pytest.mark.parametrize(
    "center, radius, output_shape, expected_shape",
    [
        ((1, 1), 1, (360, 7), (3, 2, 360, 7)),
        ((5, 8), 10, (10, 10), (3, 2, 10, 10)),
    ],
)
def test_chunk_polar_gpu(center, radius, output_shape, expected_shape):
    chunk = cp.ones((3, 2, 10, 7))
    polar_chunk = ptu._chunk_to_polar(chunk, center, radius, output_shape)
    cp.testing.assert_array_almost_equal(polar_chunk.shape, expected_shape)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from skimage import draw

from pyxem.generators import SubpixelrefinementGenerator
from pyxem.generators.subpixelrefinement_generator import (
    get_simulated_disc,
    get_experimental_square,
)
from pyxem.signals import DiffractionVectors, ElectronDiffraction2D


@pytest.fixture()
def exp_disc():
    ss, disc_radius, upsample_factor = int(60), 6, 10

    arr = np.zeros((ss, ss))
    rr, cc = draw.disk(
        (int(ss / 2) + 20, int(ss / 2) - 10), radius=disc_radius, shape=arr.shape
    )
    arr[rr, cc] = 1
    return arr


@pytest.mark.filterwarnings("ignore::UserWarning")  # various skimage warnings
def test_experimental_square_size(exp_disc):
    square = get_experimental_square(exp_disc, [17, 19], 6)
    assert square.shape[0] == int(6)
    assert square.shape[1] == int(6)


def test_failure_for_non_even_entry_to_get_simulated_disc():
    with pytest.raises(ValueError, match="'square_size' must be an even number"):
        _ = get_simulated_disc(61, 5)


def test_failure_for_non_even_errors_get_experimental_square(exp_disc):
    with pytest.raises(ValueError, match="'square_size' must be an even number"):
        _ = get_experimental_square(exp_disc, [17, 19], 7)


class Test_init_xfails:
    def test_out_of_range_vectors_numpy(self):
        """Test that putting vectors that lie outside of the
        diffraction patterns raises a ValueError"""
        vector = np.array([[1, -100]])
        dp = ElectronDiffraction2D(np.ones((20, 20)))

        with pytest.raises(ValueError):
            _ = SubpixelrefinementGenerator(dp, vector)

    def test_out_of_range_vectors_DiffractionVectors(self):
        """Test that putting vectors that lie outside of the
        diffraction patterns raises a ValueError"""
        vectors = DiffractionVectors(np.array([[1, -100]]))
        dp = ElectronDiffraction2D(np.ones((20, 20)))

        with pytest.raises(ValueError):
            _ = SubpixelrefinementGenerator(dp, vectors)

    def test_wrong_navigation_dimensions(self):
        """Tests that navigation dimensions must be appropriate too."""
        dp = ElectronDiffraction2D(np.zeros((2, 2, 8, 8)))
        vectors = DiffractionVectors(np.zeros((1, 2)))
        # Note - uses regex via re.search()
        with pytest.raises(
            ValueError,
            match=r"Vectors with shape .* must have the same navigation shape as .*",
        ):
            _ = SubpixelrefinementGenerator(dp, vectors)


class set_up_for_subpixelpeakfinders:
    def create_spot(self):
        z1, z1a = np.zeros((128, 128)), np.zeros((128, 128))
        z2, z2a = np.zeros((128, 128)), np.zeros((128, 128))

        rr, cc = draw.disk((30, 90), radius=4, shape=z1.shape)  # 30 is y!
        z1[rr, cc], z2[rr, cc] = 1, 1
        rr2, cc2 = draw.disk((100, 60), radius=4, shape=z2.shape)
        z2[rr2, cc2] = 1
        rr, cc = draw.disk((30, 90 + 3), radius=4, shape=z1.shape)  # 30 is y!
        z1a[rr, cc], z2a[rr, cc] = 1, 1
        rr2, cc2 = draw.disk((100 - 2, 60), radius=4, shape=z2.shape)
        z2a[rr2, cc2] = 1

        # marks centers for local com and local_gaussian_method
        z1[30, 90], z2[30, 90], z2[100, 60] = 2, 2, 2
        z1a[30, 93], z2a[30, 93], z2a[98, 60] = 10, 10, 10

        dp = ElectronDiffraction2D(
            np.asarray([[z1, z1a], [z2, z2a]])
        )  # this needs to be in 2x2
        return dp

    def create_Diffraction_vectors(self):
        v1 = np.array([[90 - 64, 30 - 64]])
        v2 = np.array([[90 - 64, 30 - 64], [100 - 64, 60 - 64]])
        vectors = DiffractionVectors(np.array([[v1, v1], [v2, v2]]))
        vectors.axes_manager.set_signal_dimension(0)
        return vectors


class TestSubpixelPeakFinders:
    """Tests the various peak finders have the correct x,y conventions for
    both the vectors and the shifts, in both the numpy and the DiffractionVectors
    cases as well as confirming we have avoided 'off by one' errors"""

    @pytest.fixture
    def dp(self):
        z1, z1a = np.zeros((128, 128)), np.zeros((128, 128))
        z2, z2a = np.zeros((128, 128)), np.zeros((128, 128))

        rr, cc = draw.disk((30, 90), radius=4, shape=z1.shape)  # 30 is y!
        z1[rr, cc], z2[rr, cc] = 1, 1
        rr2, cc2 = draw.disk((100, 60), radius=4, shape=z2.shape)
        z2[rr2, cc2] = 1
        rr, cc = draw.disk((30, 90 + 3), radius=4, shape=z1.shape)  # 30 is y!
        z1a[rr, cc], z2a[rr, cc] = 1, 1
        rr2, cc2 = draw.disk((100 - 2, 60), radius=4, shape=z2.shape)
        z2a[rr2, cc2] = 1

        # marks centers for local com and local_gaussian_method
        z1[30, 90], z2[30, 90], z2[100, 60] = 2, 2, 2
        z1a[30, 93], z2a[30, 93], z2a[98, 60] = 10, 10, 10

        return ElectronDiffraction2D(
            np.asarray([[z1, z1a], [z2, z2a]])
        )  # this needs to be in 2x2

    @pytest.fixture
    def diffraction_vectors(self):
        v1 = np.array([[90 - 64, 30 - 64]])
        v2 = np.array([[90 - 64, 30 - 64], [100 - 64, 60 - 64]])
        return DiffractionVectors(
            np.array([[v1, v1], [v2, v2]], dtype=object),
            ragged=True,
        )

    @pytest.fixture
    def diffraction_vectors_1_empty(self):
        v1 = np.empty((0, 2))
        v2 = np.array([[90 - 64, 30 - 64], [100 - 64, 60 - 64]])
        return DiffractionVectors(
            np.array([[v1, v1], [v2, v2]], dtype=object),
            ragged=True,
        )

    @pytest.fixture
    def sprg(self, dp, diffraction_vectors):
        return SubpixelrefinementGenerator(dp, diffraction_vectors)

    @pytest.fixture
    def sprg_empty(self, dp, diffraction_vectors_1_empty):
        return SubpixelrefinementGenerator(dp, diffraction_vectors_1_empty)

    def no_shift_case(self, s):
        error = s.data[0, 0] - np.asarray([[90 - 64, 30 - 64]])
        rms_error = np.sqrt(error[0, 0] ** 2 + error[0, 1] ** 2)
        assert rms_error < 1e-5  # perfect detection for this trivial case

    def x_shift_case(self, s):
        error = s.data[0, 1] - np.asarray([[93 - 64, 30 - 64]])
        rms_error = np.sqrt(error[0, 0] ** 2 + error[0, 1] ** 2)
        assert rms_error < 0.5  # correct to within a pixel

    @pytest.mark.skip(reason="Broken downstream in scikit image")
    def test_assertioned_xc(self, sprg):
        subpixelsfound = sprg.conventional_xc(12, 4, 8)
        self.no_shift_case(subpixelsfound)
        self.x_shift_case(subpixelsfound)

    def test_assertioned_com(self, sprg):
        vector_refine = sprg.center_of_mass_method(12)
        self.no_shift_case(vector_refine)
        self.x_shift_case(vector_refine)

    def test_assertioned_com_empty(self, sprg_empty):
        vector_refine = sprg_empty.center_of_mass_method(12)
        assert isinstance(vector_refine, DiffractionVectors)

    def test_failure_out_of_bounds(self, dp):
        v = [[1000, 100]]
        vectors = np.array([[v, v], [v, v]])
        with pytest.raises(ValueError):
            SubpixelrefinementGenerator(dp, vectors)

    def test_log(self, sprg):
        with pytest.raises(
            NotImplementedError,
            match="This functionality was removed in v.0.13.0",
        ):
            _ = sprg.local_gaussian_method(12)


def test_xy_errors_in_conventional_xc_method_as_per_issue_490():
    """This was the MWE example code for the issue"""
    dp = get_simulated_disc(100, 20)
    # translate y by +4
    shifted = np.pad(dp, ((0, 4), (0, 0)), "constant")[4:].reshape(1, 1, *dp.shape)
    signal = ElectronDiffraction2D(shifted)
    spg = SubpixelrefinementGenerator(signal, np.array([[0, 0]]))
    peaks = spg.conventional_xc(100, 20, 1).data[0][0][0]  # as quoted in the issue
    np.testing.assert_allclose([0, -4], peaks)
    """ we also test com method for clarity """
    peaks = spg.center_of_mass_method(60).data[0][0][0]
    np.testing.assert_allclose([0, -4], peaks, atol=1.5)
    """ we also test reference_xc """
    peaks = spg.reference_xc(100, dp, 1).data[0][0][0]  # as quoted in the issue
    np.testing.assert_allclose([0, -4], peaks)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from pyxem.signals import (
    DiffractionVariance2D,
    ImageVariance,
    DiffractionVariance1D,
)


class TestDiffractionVariance:
    def test_get_diffraction_variance_signal(self, diffraction_pattern):
        difvar = DiffractionVariance2D(diffraction_pattern)
        assert isinstance(difvar, DiffractionVariance2D)

    def test_1d_azimuthal_integration(self):
        var = DiffractionVariance2D(
            data=np.ones(
                (
                    3,
                    3,
                    3,
                    3,
                )
            )
        )
        var.unit = "2th_rad"
        integration = var.get_azimuthal_integral1d(npt=10)
        assert isinstance(integration, DiffractionVariance1D)


class TestImageVariance:
    def test_get_image_variance_signal(self, diffraction_pattern):
        imvar = ImageVariance(diffraction_pattern)
        assert isinstance(imvar, ImageVariance)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import dask.array as da
from scipy.ndimage import gaussian_filter
from scipy.signal import convolve2d
from skimage import morphology
from skimage.draw import polygon

from hyperspy.components1d import Gaussian

import pyxem.data.dummy_data.make_diffraction_test_data as mdtd
from pyxem.signals import Diffraction2D, LazyDiffraction2D, BeamShift


def get_disk_shift_simple_test_signal(lazy=False):
    """Get HyperSpy 2D signal with 2D navigation dimensions.

    Useful for exploring the effects of dscan.

    Probe size x/y (20, 20), and image size x/y (50, 50).
    Disk moves from 22-28 x/y.

    Parameters
    ----------
    lazy : bool, default False

    Returns
    -------
    disk_shift_signal : Diffraction2D

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_disk_shift_simple_test_signal()
    >>> s.plot()

    Load as lazy

    >>> s = pxm.data.dummy_data.get_disk_shift_simple_test_signal(lazy=True)

    """
    disk_x, disk_y = np.mgrid[22:28:20j, 22:28:20j]
    s = mdtd.generate_4d_data(
        probe_size_x=20,
        probe_size_y=20,
        image_size_x=50,
        image_size_y=50,
        disk_x=disk_x,
        disk_y=disk_y,
        disk_r=2,
        ring_x=None,
        add_noise=True,
        lazy=lazy,
    )
    return s


def get_holz_simple_test_signal(lazy=False):
    """Get HyperSpy 2D signal with 2D navigation dimensions for HOLZ testing.

    Probe size x/y (20, 20), and image size x/y (50, 50).
    Contains a disk and a ring. The disk stays at x, y = 25, 25, with radius 2.
    The ring has a radius of 20, and moves from x, y = 24-26, 24-26.

    Parameters
    ----------
    lazy : bool, default False

    Returns
    -------
    holz_signal : Diffraction2D

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_holz_simple_test_signal()
    >>> s.plot()

    Load as lazy

    >>> s = pxm.data.dummy_data.get_holz_simple_test_signal(lazy=True)

    """
    ring_x, ring_y = np.mgrid[24:26:20j, 24:26:20j]
    s = mdtd.generate_4d_data(
        probe_size_x=20,
        probe_size_y=20,
        image_size_x=50,
        image_size_y=50,
        disk_x=25,
        disk_y=25,
        disk_r=2,
        disk_I=20,
        ring_x=ring_x,
        ring_y=ring_y,
        ring_r=15,
        ring_I=10,
        add_noise=True,
        lazy=lazy,
    )
    return s


def get_holz_heterostructure_test_signal(lazy=False, probe_size_x=40, probe_size_y=40):
    """Get HyperSpy 2D signal with 2D navigation dimensions for HOLZ testing.

    The centre, radius and intensity of the ring varies as a function of probe
    position. The disk centre position varies as a function of probe position.

    Parameters
    ----------
    lazy : bool, default False

    Returns
    -------
    holz_signal : HyperSpy 2D signal

    Example
    -------
    >>> s = pxm.data.dummy_data.get_holz_heterostructure_test_signal()
    >>> s.plot()

    Load as lazy

    >>> s = pxm.data.dummy_data.get_holz_heterostructure_test_signal(lazy=True)

    """
    px, py = np.mgrid[0:probe_size_x:1, 0:probe_size_y:1]

    centerx = probe_size_x // 2
    centery = probe_size_x // 2

    x, y = np.mgrid[
        centerx - 4 : centerx - 2 : probe_size_x * 1j,
        centery + 1 : centery + 3 : probe_size_y * 1j,
    ]
    disk_r = 10
    disk_I = np.ones_like(x) * 100 + np.random.random() * 20
    g_r = Gaussian(A=20, centre=25, sigma=5)
    ring_r = np.ones_like(x) * 30 + g_r.function(py)
    g_I = Gaussian(A=30, centre=25, sigma=3)
    ring_I = np.ones_like(x) * 20 + g_I.function(py)
    s = mdtd.generate_4d_data(
        probe_size_x=probe_size_x,
        probe_size_y=probe_size_y,
        image_size_x=80,
        image_size_y=80,
        disk_x=x,
        disk_y=y,
        disk_r=disk_r,
        disk_I=disk_I,
        ring_x=x,
        ring_y=y,
        ring_r=ring_r,
        ring_I=ring_I,
        add_noise=True,
        lazy=lazy,
    )
    return s


def get_single_ring_diffraction_signal():
    """Get HyperSpy 2D signal with a single ring with centre position.

    The ring has a centre at x=105 and y=67, and radius=40.
    """
    data = mdtd.MakeTestData(size_x=200, size_y=150, default=False, blur=True)
    x, y = 105, 67
    data.add_ring(x, y, r=40)
    s = data.signal
    s.axes_manager[0].offset, s.axes_manager[1].offset = -x, -y
    return s


def get_dead_pixel_signal(lazy=False):
    """Get Diffraction2D signal with a disk in the middle.

    Has 4 pixels with value equal to 0, to simulate dead pixels.

    Example
    -------
    >>> s = pxm.data.dummy_data.get_dead_pixel_signal()

    Lazy signal

    >>> s_lazy = pxm.data.dummy_data.get_dead_pixel_signal(lazy=True)

    """
    data = mdtd.MakeTestData(size_x=128, size_y=128, default=False, blur=True)
    data.add_disk(64, 64, r=30, intensity=10000)
    s = data.signal
    s.change_dtype("int64")
    s.data += gaussian_filter(s.data, sigma=50)
    s.data[61, 73] = 0
    s.data[46, 53] = 0
    s.data[88, 88] = 0
    s.data[112, 20] = 0
    if lazy:
        s = LazyDiffraction2D(s)
        s.data = da.from_array(s.data, chunks=(64, 64))
    else:
        s = Diffraction2D(s)
    return s


def get_hot_pixel_signal(lazy=False):
    """Get Diffraction2D signal with a disk in the middle.

    Has 4 pixels with value equal to 50000, to simulate hot pixels.

    Example
    -------
    >>> s = pxm.data.dummy_data.get_hot_pixel_signal()

    Lazy signal

    >>> s_lazy = pxm.data.dummy_data.get_hot_pixel_signal(lazy=True)

    """
    data = mdtd.MakeTestData(size_x=128, size_y=128, default=False, blur=True)
    data.add_disk(64, 64, r=30, intensity=10000)
    s = data.signal
    s.change_dtype("int64")
    s.data += gaussian_filter(s.data, sigma=50)
    s.data[76, 4] = 50000
    s.data[12, 102] = 50000
    s.data[32, 10] = 50000
    s.data[120, 61] = 50000
    if lazy:
        s = LazyDiffraction2D(s)
        s.data = da.from_array(s.data, chunks=(64, 64))
    else:
        s = Diffraction2D(s)
    return s


def get_simple_beam_shift_signal():
    """Get a simple BeamShift signal with a zero point in the centre.

    Example
    -------
    >>> s = pxm.data.dummy_data.get_simple_beam_shift_signal()

    """
    temp_data = np.mgrid[-5:5:100j, -5:5:100j]
    data = np.stack((temp_data[0], temp_data[1]), -1)
    s = BeamShift(data)
    return s


def get_stripe_pattern_beam_shift_signal():
    """Get a BeamShift signal with a stripe pattern.

    The stripe pattern only has an x-component, with alternating left/right
    directions. There is a small a net moment in the positive x-direction
    (leftwards).

    Returns
    -------
    stripe_beam_shift_signal : BeamShift

    Example
    -------
    >>> s = pxm.data.dummy_data.get_stripe_pattern_beam_shift_signal()

    """
    data = np.zeros((100, 50, 2))
    for i in range(10, 90, 20):
        data[i : i + 10, 10:40, 0] = 1.1
        data[i + 10 : i + 20, 10:40, 0] = -1
    s = BeamShift(data)
    return s


def get_magnetic_square_beam_shift_signal(add_ramp=False):
    """Get a BeamShift signal resembling a Landau domain.

    Parameters
    ----------
    add_ramp : bool, default False
        If True, will add a ramp in the beam shift across the image
        to emulate the effects of d-scan.

    Returns
    -------
    square_beam_shift_signal : BeamShift

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_magnetic_square_beam_shift_signal()
    >>> s.plot()

    Adding a ramp

    >>> s = pxm.data.dummy_data.get_magnetic_square_beam_shift_signal(add_ramp=True)
    >>> s.plot()

    """
    imX, imY = 300, 300
    data_y, data_x = np.random.normal(loc=0.1, size=(2, imY, imX))
    x, y = np.mgrid[-150 : 150 : imY * 1j, -150 : 150 : imX * 1j]
    t = np.arctan2(x, y) % (2 * np.pi)
    mask_xy = (x < 100) * (x > -100) * (y < 100) * (y > -100)
    mask0 = (t > np.pi / 4) * (t < 3 * np.pi / 4) * mask_xy
    mask1 = (t > 3 * np.pi / 4) * (t < 5 * np.pi / 4) * mask_xy
    mask2 = (t > 5 * np.pi / 4) * (t < 7 * np.pi / 4) * mask_xy
    mask3 = ((t > 7 * np.pi / 4) + (t < np.pi / 4)) * mask_xy
    data_y[mask0] -= np.random.normal(loc=3, scale=0.1, size=(imY, imX))[mask0]
    data_x[mask1] -= np.random.normal(loc=3, scale=0.1, size=(imY, imX))[mask1]
    data_y[mask2] += np.random.normal(loc=3, scale=0.1, size=(imY, imX))[mask2]
    data_x[mask3] += np.random.normal(loc=3, scale=0.1, size=(imY, imX))[mask3]
    if add_ramp:
        ramp_y, ramp_x = np.mgrid[18 : 28 : imY * 1j, -5.3 : 1.2 : imX * 1j]
        data_x += ramp_x
        data_y += ramp_y
    data = np.stack((data_y, data_x), axis=-1)
    s = BeamShift(data)
    s.axes_manager.navigation_axes[0].name = "Probe position x"
    s.axes_manager.navigation_axes[1].name = "Probe position y"
    s.axes_manager.navigation_axes[0].units = "nm"
    s.axes_manager.navigation_axes[1].units = "nm"
    return s


def get_fem_signal(lazy=False):
    """Get a 2D signal approximating a fluctuation electron microscopy (FEM) dataset.

    Parameters
    ----------
    lazy : bool, default False
        If True, resulting signal will be lazy.

    Returns
    -------
    fem_signal : Diffraction2D

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_fem_signal()
    >>> s.plot()

    """
    radii1 = 20 * np.random.randint(0, 2, size=(10, 10))
    intensities1 = np.random.randint(0, 30, size=(10, 10))
    radii2 = 35 * np.random.randint(0, 2, size=(10, 10))
    intensities2 = np.random.randint(0, 30, size=(10, 10))

    test1 = mdtd.generate_4d_data(
        probe_size_x=10,
        probe_size_y=10,
        image_size_x=100,
        image_size_y=100,
        disk_x=50,
        disk_y=50,
        disk_r=5,
        disk_I=100,
        ring_x=50,
        ring_y=50,
        ring_r=radii1,
        ring_I=intensities1,
        ring_lw=0,
        blur=True,
        blur_sigma=1,
        downscale=True,
        add_noise=True,
        show_progressbar=False,
        lazy=lazy,
    )

    test2 = mdtd.generate_4d_data(
        probe_size_x=10,
        probe_size_y=10,
        image_size_x=100,
        image_size_y=100,
        disk_x=50,
        disk_y=50,
        disk_r=5,
        disk_I=100,
        ring_x=50,
        ring_y=50,
        ring_r=radii2,
        ring_I=intensities2,
        ring_lw=0,
        blur=True,
        blur_sigma=1,
        downscale=True,
        add_noise=True,
        show_progressbar=False,
        lazy=lazy,
    )

    fem_signal = test1 + test2

    fem_signal.axes_manager.navigation_axes[0].name = "Probe position x"
    fem_signal.axes_manager.navigation_axes[0].units = "nm"
    fem_signal.axes_manager.navigation_axes[0].scale = 1.0
    fem_signal.axes_manager.navigation_axes[0].offset = 0

    fem_signal.axes_manager.navigation_axes[1].name = "Probe position y"
    fem_signal.axes_manager.navigation_axes[1].units = "nm"
    fem_signal.axes_manager.navigation_axes[1].scale = 1.0
    fem_signal.axes_manager.navigation_axes[1].offset = 0

    fem_signal.axes_manager.signal_axes[0].name = "Signal x"
    fem_signal.axes_manager.signal_axes[0].units = "mrads"
    fem_signal.axes_manager.signal_axes[0].scale = 0.25
    fem_signal.axes_manager.signal_axes[0].offset = 0

    fem_signal.axes_manager.signal_axes[1].name = "Signal y"
    fem_signal.axes_manager.signal_axes[1].units = "mrads"
    fem_signal.axes_manager.signal_axes[1].scale = 0.25
    fem_signal.axes_manager.signal_axes[1].offset = 0

    return fem_signal


def get_simple_fem_signal(lazy=False):
    """2D signal approximating a very small fluctuation electron microscopy (FEM) dataset.

    Parameters
    ----------
    lazy : bool, default False
        If True, resulting signal will be lazy.

    Returns
    -------
    fem_signal : Diffraction2D

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_simple_fem_signal()
    >>> s.plot()

    """

    radii1 = 10 * np.random.randint(0, 2, size=(2, 2))
    intensities1 = np.random.randint(0, 5, size=(2, 2))

    radii2 = 20 * np.random.randint(0, 2, size=(2, 2))
    intensities2 = np.random.randint(0, 15, size=(2, 2))

    test1 = mdtd.generate_4d_data(
        probe_size_x=2,
        probe_size_y=2,
        image_size_x=50,
        image_size_y=50,
        disk_x=25,
        disk_y=25,
        disk_r=5,
        disk_I=100,
        ring_x=25,
        ring_y=25,
        ring_r=radii1,
        ring_I=intensities1,
        ring_lw=0,
        blur=True,
        blur_sigma=1,
        downscale=True,
        add_noise=True,
        show_progressbar=False,
        lazy=lazy,
    )

    test2 = mdtd.generate_4d_data(
        probe_size_x=2,
        probe_size_y=2,
        image_size_x=50,
        image_size_y=50,
        disk_x=25,
        disk_y=25,
        disk_r=5,
        disk_I=100,
        ring_x=25,
        ring_y=25,
        ring_r=radii2,
        ring_I=intensities2,
        ring_lw=0,
        blur=True,
        blur_sigma=1,
        downscale=True,
        add_noise=True,
        show_progressbar=False,
        lazy=lazy,
    )

    fem_signal = test1 + test2

    fem_signal.axes_manager.navigation_axes[0].name = "Probe position x"
    fem_signal.axes_manager.navigation_axes[0].units = "nm"
    fem_signal.axes_manager.navigation_axes[0].scale = 1.0
    fem_signal.axes_manager.navigation_axes[0].offset = 0

    fem_signal.axes_manager.navigation_axes[1].name = "Probe position y"
    fem_signal.axes_manager.navigation_axes[1].units = "nm"
    fem_signal.axes_manager.navigation_axes[1].scale = 1.0
    fem_signal.axes_manager.navigation_axes[1].offset = 0

    fem_signal.axes_manager.signal_axes[0].name = "Signal x"
    fem_signal.axes_manager.signal_axes[0].units = "mrads"
    fem_signal.axes_manager.signal_axes[0].scale = 0.25
    fem_signal.axes_manager.signal_axes[0].offset = 0

    fem_signal.axes_manager.signal_axes[1].name = "Signal y"
    fem_signal.axes_manager.signal_axes[1].units = "mrads"
    fem_signal.axes_manager.signal_axes[1].scale = 0.25
    fem_signal.axes_manager.signal_axes[1].offset = 0

    return fem_signal


def get_generic_fem_signal(probe_x=2, probe_y=2, image_x=50, image_y=50, lazy=False):
    """2D signal approximating a fluctuation electron microscopy (FEM) dataset with user defined dimensions.

    Parameters
    ----------
    probe_x : int, default 2
        Horizontal dimension of the navigation axes
    probe_y : int, default 2
        Vertical dimension of the navigation axes
    image_x : int, default 2
        Horizontal dimension of the signal axes
    image_y : int, default 2
        Vertical dimension of the signal axes
    lazy : bool, default False
        If True, resulting signal will be lazy.

    Returns
    -------
    fem_signal : Diffraction2D

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_generic_fem_signal(probe_x=5, probe_y=10,
    ...     image_x=25, image_y=30, lazy=False)
    >>> s.plot()

    """
    image_center = [int(image_x / 2), int(image_y / 2)]

    radii1 = 10 * np.random.randint(0, 2, size=(probe_y, probe_x))
    intensities1 = np.random.randint(0, 5, size=(probe_y, probe_x))

    radii2 = 20 * np.random.randint(0, 2, size=(probe_y, probe_x))
    intensities2 = np.random.randint(0, 15, size=(probe_y, probe_x))

    test1 = mdtd.generate_4d_data(
        probe_size_x=probe_x,
        probe_size_y=probe_y,
        image_size_x=image_x,
        image_size_y=image_y,
        disk_x=image_center[0],
        disk_y=image_center[1],
        disk_r=5,
        disk_I=100,
        ring_x=image_center[0],
        ring_y=image_center[1],
        ring_r=radii1,
        ring_I=intensities1,
        ring_lw=0,
        blur=True,
        blur_sigma=1,
        downscale=True,
        add_noise=True,
        show_progressbar=False,
        lazy=lazy,
    )

    test2 = mdtd.generate_4d_data(
        probe_size_x=probe_x,
        probe_size_y=probe_y,
        image_size_x=image_x,
        image_size_y=image_y,
        disk_x=image_center[0],
        disk_y=image_center[1],
        disk_r=5,
        disk_I=100,
        ring_x=image_center[0],
        ring_y=image_center[1],
        ring_r=radii2,
        ring_I=intensities2,
        ring_lw=0,
        blur=True,
        blur_sigma=1,
        downscale=True,
        add_noise=True,
        show_progressbar=False,
        lazy=lazy,
    )

    fem_signal = test1 + test2

    fem_signal.axes_manager.navigation_axes[0].name = "Probe position x"
    fem_signal.axes_manager.navigation_axes[0].units = "nm"
    fem_signal.axes_manager.navigation_axes[0].scale = 1.0
    fem_signal.axes_manager.navigation_axes[0].offset = 0

    fem_signal.axes_manager.navigation_axes[1].name = "Probe position y"
    fem_signal.axes_manager.navigation_axes[1].units = "nm"
    fem_signal.axes_manager.navigation_axes[1].scale = 1.0
    fem_signal.axes_manager.navigation_axes[1].offset = 0

    fem_signal.axes_manager.signal_axes[0].name = "Signal x"
    fem_signal.axes_manager.signal_axes[0].units = "mrads"
    fem_signal.axes_manager.signal_axes[0].scale = 0.25
    fem_signal.axes_manager.signal_axes[0].offset = 0

    fem_signal.axes_manager.signal_axes[1].name = "Signal y"
    fem_signal.axes_manager.signal_axes[1].units = "mrads"
    fem_signal.axes_manager.signal_axes[1].scale = 0.25
    fem_signal.axes_manager.signal_axes[1].offset = 0
    return fem_signal


def get_cbed_signal():
    """Get artificial pixelated STEM signal similar to CBED data.

    Returns
    -------
    cbed_signal : Diffraction2D

    Example
    -------
    >>> s = pxm.data.dummy_data.get_cbed_signal()
    >>> s.plot()

    """
    data = np.zeros(shape=(10, 10, 100, 100), dtype=np.uint16)
    diff_point_list = [
        [50, 50, 1000],
        [25, 25, 500],
        [50, 25, 500],
        [25, 50, 500],
        [75, 75, 500],
        [75, 50, 500],
        [50, 75, 500],
        [75, 25, 500],
        [25, 75, 500],
    ]
    disk = morphology.disk(5, np.uint16)
    for x, y, intensity in diff_point_list:
        for ix, iy in np.ndindex(data.shape[:2]):
            temp_x = x
            temp_y = y
            temp_intensity = intensity
            if iy < 5:
                if y < 40:
                    temp_y = y + 2
                    temp_intensity = intensity + 20
                elif y > 60:
                    temp_y = y - 2

            data[iy, ix, temp_y, temp_x] = temp_intensity
    for ix, iy in np.ndindex(data.shape[:2]):
        noise = np.random.randint(10, size=(100, 100))
        image = convolve2d(data[iy, ix], disk, mode="same") + noise
        data[iy, ix] = gaussian_filter(image, 1.5)
    s_cbed = Diffraction2D(data)
    return s_cbed


def get_simple_ellipse_signal_peak_array(seed=None):
    """Get a signal and peak array of an ellipse.

    Returns
    -------
    signal, peak_array : HyperSpy Signal2D, NumPy array

    Examples
    --------
    >>> s, peak_array = pxm.data.dummy_data.get_simple_ellipse_signal_peak_array()
    >>> s.add_peak_array_as_markers(peak_array, color='blue', size=30)

    """
    rng = np.random.default_rng(seed=seed)
    xc = rng.integers(95, 105, size=(4, 5))
    yc = rng.integers(95, 105, size=(4, 5))
    semi0 = rng.integers(55, 60, size=(4, 5))
    semi1 = rng.integers(75, 80, size=(4, 5))
    rot = rng.random(size=(4, 5)) * np.pi
    peak_array = mdtd._make_4d_peak_array_test_data(xc, yc, semi0, semi1, rot)
    s = Diffraction2D(np.zeros((4, 5, 200, 200)))
    return s, peak_array


def get_nanobeam_electron_diffraction_signal():
    """Get a signal emulating a NBED dataset.

    Returns
    -------
    signal : Diffraction2D

    Example
    -------
    >>> s = pxm.data.dummy_data.get_nanobeam_electron_diffraction_signal()
    >>> s.plot()

    """
    di0 = mdtd.DiffractionTestImage(intensity_noise=False)
    di0.add_disk(x=128, y=128, intensity=10.0)
    di0.add_cubic_disks(vx=20, vy=20, intensity=2.0, n=5)
    di0.add_background_lorentz(intensity=50, width=30)

    di1 = di0.copy()
    di1.rotation = 10
    di2 = di0.copy()
    di2.rotation = -10

    position_array0 = np.zeros((50, 50), dtype=bool)
    r = np.array([15, 15, 0, 0])
    c = np.array([0, 15, 31, 0])
    rr, cc = polygon(r, c)
    position_array0[rr, cc] = True

    r = np.array([10, 19, 29, 40])
    c = np.array([49, 35, 35, 49])
    rr, cc = polygon(r, c)
    position_array0[rr, cc] = True

    position_array1 = np.zeros((50, 50), dtype=bool)
    r = np.array([32, 41, 41, 49, 49])
    c = np.array([0, 18, 49, 49, 0])
    rr, cc = polygon(r, c)
    position_array1[rr, cc] = True

    position_array2 = np.invert(np.bitwise_or(position_array0, position_array1))

    dtd = mdtd.DiffractionTestDataset(50, 50, 256, 256)
    dtd.add_diffraction_image(di0, position_array0)
    dtd.add_diffraction_image(di1, position_array1)
    dtd.add_diffraction_image(di2, position_array2)

    s = dtd.get_signal()
    return s



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import dask.array as da
import numpy as np

from pyxem.signals import (
    DPCSignal1D,
    DPCSignal2D,
    LazyDPCSignal1D,
    LazyDPCSignal2D,
)


class TestDpcSignal1dCreate:
    def test_create(self):
        data = np.ones(shape=(2, 10))
        DPCSignal1D(data)

    def test_create_lazy(self):
        data = da.ones(shape=(2, 10))
        LazyDPCSignal1D(data)


class TestDpcSignal2dCreate:
    def test_create(self):
        data = np.ones(shape=(2, 10, 10))
        DPCSignal2D(data)
        with pytest.raises(ValueError):
            DPCSignal2D(np.zeros(10))

    def test_create_lazy(self):
        data = da.ones(shape=(2, 10, 10))
        LazyDPCSignal2D(data)


class TestToBeamShift:
    def test_dpc_signal2d(self):
        probe_x, probe_y = 100, 200
        s = DPCSignal2D(np.zeros((2, probe_y, probe_x)))
        s_beam_shift = s.to_beamshift()
        assert s_beam_shift.axes_manager.shape == (probe_x, probe_y, 2)

    def test_dpc_signal1d(self):
        probe_x = 200
        s = DPCSignal1D(np.zeros((2, probe_x)))
        s_beam_shift = s.to_beamshift()
        assert s_beam_shift.axes_manager.shape == (probe_x, 2)

    def test_lazy_dpc_signal2d(self):
        probe_x, probe_y = 100, 200
        s = LazyDPCSignal2D(da.zeros((2, probe_y, probe_x)))
        s_beam_shift = s.to_beamshift()
        assert s_beam_shift._lazy
        assert s_beam_shift.axes_manager.shape == (probe_x, probe_y, 2)

    def test_lazy_dpc_signal1d(self):
        probe_x = 200
        s = LazyDPCSignal1D(da.zeros((2, probe_x)))
        s_beam_shift = s.to_beamshift()
        assert s_beam_shift._lazy
        assert s_beam_shift.axes_manager.shape == (probe_x, 2)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import warnings

import numpy as np
import pytest

from pyxem.utils._deprecated import deprecated, deprecated_argument
from pyxem.common import VisibleDeprecationWarning


class TestDeprecationWarning:
    def test_deprecation_since(self):
        """Ensure functions decorated with the custom deprecated
        decorator returns desired output, raises a desired warning, and
        gets the desired additions to their docstring.
        """

        @deprecated(since=0.7, alternative="bar", removal=0.8)
        def foo(n):
            """Some docstring."""
            return n + 1

        with pytest.warns(VisibleDeprecationWarning) as record:
            assert foo(4) == 5
        desired_msg = (
            "Function `foo()` is deprecated and will be removed in version 0.8. Use "
            "`bar()` instead."
        )
        assert str(record[0].message) == desired_msg
        assert foo.__doc__ == (
            "[*Deprecated*] Some docstring.\n\n"
            "Notes\n-----\n"
            ".. deprecated:: 0.7\n"
            f"   {desired_msg}"
        )

        @deprecated(since=1.9)
        def foo2(n):
            """Another docstring.
            Notes
            -----
            Some existing notes.
            """
            return n + 2

        with pytest.warns(VisibleDeprecationWarning) as record2:
            assert foo2(4) == 6
        desired_msg2 = "Function `foo2()` is deprecated."
        assert str(record2[0].message) == desired_msg2
        assert foo2.__doc__ == (
            "[*Deprecated*] Another docstring."
            "\nNotes\n-----\n"
            "Some existing notes.\n\n"
            ".. deprecated:: 1.9\n"
            f"   {desired_msg2}"
        )

    def test_deprecation_no_old_doc(self):
        @deprecated(since=0.7, alternative="bar", removal=0.8)
        def foo(n):
            return n + 1

        with pytest.warns(VisibleDeprecationWarning) as record:
            assert foo(4) == 5
        desired_msg = (
            "Function `foo()` is deprecated and will be removed in version 0.8. Use "
            "`bar()` instead."
        )
        assert str(record[0].message) == desired_msg
        assert foo.__doc__ == (
            "[*Deprecated*] \n"
            "\nNotes\n-----\n"
            ".. deprecated:: 0.7\n"
            f"   {desired_msg}"
        )


class TestDeprecateArgument:
    def test_deprecate_argument(self):
        """Functions decorated with the custom `deprecated_argument`
        decorator returns desired output and raises a desired warning
        only if the argument is passed.
        """

        class Foo:
            @deprecated_argument(name="a", since="1.3", removal="1.4")
            def bar_arg(self, **kwargs):
                return kwargs

            @deprecated_argument(name="a", since="1.3", removal="1.4", alternative="b")
            def bar_arg_alt(self, **kwargs):
                return kwargs

        my_foo = Foo()

        # Does not warn
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            assert my_foo.bar_arg(b=1) == {"b": 1}

        # Warns
        with pytest.warns(VisibleDeprecationWarning) as record2:
            assert my_foo.bar_arg(a=2) == {"a": 2}
        assert str(record2[0].message) == (
            r"Argument `a` is deprecated and will be removed in version 1.4. "
            r"To avoid this warning, please do not use `a`. See the documentation of "
            r"`bar_arg()` for more details."
        )

        # Warns with alternative
        with pytest.warns(VisibleDeprecationWarning) as record3:
            assert my_foo.bar_arg_alt(a=3) == {"b": 3}
        assert str(record3[0].message) == (
            r"Argument `a` is deprecated and will be removed in version 1.4. "
            r"To avoid this warning, please do not use `a`. Use `b` instead. See the "
            r"documentation of `bar_arg_alt()` for more details."
        )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from numpy.random import randint
import dask.array as da
from skimage import morphology

from hyperspy.signals import Signal2D, BaseSignal

import pyxem.signals
from pyxem.signals import Diffraction2D, LazyDiffraction2D
import pyxem.utils.ransac_ellipse_tools as ret
from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
import pyxem.data.dummy_data as dd


@pytest.mark.slow
class TestDiffraction2D:
    def test_create(self):
        array0 = np.zeros(shape=(10, 10, 10, 10))
        s0 = Diffraction2D(array0)
        assert array0.shape == s0.axes_manager.shape

        # This should fail due to Diffraction2D inheriting
        # signal2D, i.e. the data has to be at least
        # 2-dimensions
        with pytest.raises(ValueError):
            Diffraction2D(np.zeros(10))

        array1 = np.zeros(shape=(10, 10))
        s1 = Diffraction2D(array1)
        assert array1.shape == s1.axes_manager.shape


@pytest.mark.slow
class TestPlotting:
    def test_simple_plot(self):
        s = Diffraction2D(np.zeros(shape=(3, 4, 6, 10)))
        s.plot()

    def test_navigation_signal_plot(self):
        s = Diffraction2D(np.zeros(shape=(3, 4, 6, 10)))
        s_nav = Signal2D(np.zeros((3, 4)))
        s.navigation_signal = s_nav
        s.plot()

    def test_navigation_signal_plot_argument(self):
        s = Diffraction2D(np.zeros(shape=(3, 4, 6, 10)))
        s_nav = Signal2D(np.zeros((3, 4)))
        s.plot(navigator=s_nav)


class TestDiffraction2DFlipDiffraction:
    def test_flip_x(self):
        array = np.zeros(shape=(3, 4, 6, 10))
        array[:, :, :, 5:] = 1
        s = Diffraction2D(array)
        assert (s.data[:, :, :, 5:] == 1).all()
        s_flip = s.flip_diffraction_x()
        s.metadata.General.title = "Test"
        assert s_flip.metadata == s.metadata
        assert s_flip.metadata.General.title == "Test"
        assert (s_flip.data[:, :, :, 5:] == 0).all()
        assert (s_flip.data[:, :, :, :5] == 1).all()

    def test_flip_y(self):
        array = np.zeros(shape=(3, 4, 6, 10))
        array[:, :, 3:, :] = 1
        s = Diffraction2D(array)
        assert (s.data[:, :, 3:, :] == 1).all()
        s.metadata.General.title = "Test"
        s_flip = s.flip_diffraction_y()
        assert s_flip.metadata == s.metadata
        assert s_flip.metadata.General.title == "Test"
        assert (s_flip.data[:, :, 3:, :] == 0).all()
        assert (s_flip.data[:, :, :3, :] == 1).all()


class TestAddPeakArrayAsMarkers:
    def test_simple(self):
        s = Diffraction2D(np.zeros((2, 3, 100, 100)))
        peak_array = np.empty((2, 3), dtype=object)
        for index in np.ndindex(peak_array.shape):
            islice = np.s_[index]
            peak_array[islice] = np.random.randint(20, 80, (100, 2))
        s.add_peak_array_as_markers(peak_array)

    def test_color(self):
        s = Diffraction2D(np.zeros((2, 3, 100, 100)))
        peak_array = np.empty((2, 3), dtype=object)
        for index in np.ndindex(peak_array.shape):
            islice = np.s_[index]
            peak_array[islice] = np.random.randint(20, 80, (100, 2))
        s.add_peak_array_as_markers(peak_array, color=("blue",))
        marker0 = list(s.metadata.Markers)[0][1]
        assert marker0.kwargs["color"] == ("blue",)

    def test_peak_array(self):
        s = Diffraction2D(np.zeros((2, 3, 100, 100)))
        peak_array = np.empty((3, 2), dtype=object)
        for index in np.ndindex(peak_array.shape):
            islice = np.s_[index]
            peak_array[islice] = np.random.randint(20, 80, (100, 2))
        pk = BaseSignal(peak_array)
        s.add_peak_array_as_markers(pk)

    def test_peak_array_error(self):
        s = Diffraction2D(np.zeros((2, 3, 100, 100)))
        with pytest.raises(TypeError):
            s.add_peak_array_as_markers([1, 2, 3, 4])

    def test_size(self):
        s = Diffraction2D(np.zeros((2, 3, 100, 100)))
        peak_array = np.empty((2, 3), dtype=object)
        for index in np.ndindex(peak_array.shape):
            islice = np.s_[index]
            peak_array[islice] = np.random.randint(20, 80, (100, 2))
        s.add_peak_array_as_markers(peak_array, sizes=(13,))
        marker = list(s.metadata.Markers)[0][1]
        assert marker.kwargs["sizes"] == (13,)

    def test_3d_nav_dims(self):
        s = Diffraction2D(np.zeros((2, 3, 4, 100, 100)))
        peak_array = np.empty((2, 3, 4), dtype=object)
        for index in np.ndindex(peak_array.shape):
            islice = np.s_[index]
            peak_array[islice] = np.random.randint(20, 80, (100, 2))
        s.add_peak_array_as_markers(peak_array)
        marker = list(s.metadata.Markers)[0][1]
        assert marker.kwargs["offsets"][()].shape == (4, 3, 2)

    def test_1d_nav_dims(self):
        nav_dim = 3
        s = Diffraction2D(np.zeros((nav_dim, 100, 100)))
        peak_array = np.empty(nav_dim, dtype=object)
        for index in np.ndindex(peak_array.shape):
            islice = np.s_[index]
            peak_array[islice] = np.random.randint(20, 80, (100, 2))
        s.add_peak_array_as_markers(peak_array)
        marker = list(s.metadata.Markers)[0][1]
        assert marker.kwargs["offsets"].shape == (3,)

    def test_0d_nav_dims(self):
        s = Diffraction2D(np.zeros((100, 100)))
        peak_array = np.random.randint(20, 80, size=(100, 2))
        s.add_peak_array_as_markers(peak_array)
        marker = list(s.metadata.Markers)[0][1]
        assert marker.kwargs["offsets"].dtype != object


class TestAddEllipseArrayAsMarkers:
    def test_simple(self):
        s, parray = dd.get_simple_ellipse_signal_peak_array(seed=15)
        ellipse_array, inlier_array = ret.get_ellipse_model_ransac(
            parray,
            xf=95,
            yf=95,
            rf_lim=20,
            semi_len_min=40,
            semi_len_max=100,
            semi_len_ratio_lim=5,
            max_trials=50,
        )
        s.add_ellipse_array_as_markers(
            ellipse_array, inlier_array=inlier_array.T, peak_array=parray.T
        )

    @pytest.mark.flaky(reruns=2)
    def test_only_ellipse_array(self):
        s, parray = dd.get_simple_ellipse_signal_peak_array(seed=15)
        s1 = s.deepcopy()
        ellipse_array, inlier_array = ret.get_ellipse_model_ransac(
            parray,
            xf=95,
            yf=95,
            rf_lim=20,
            semi_len_min=40,
            semi_len_max=100,
            semi_len_ratio_lim=5,
            max_trials=50,
        )
        s.add_ellipse_array_as_markers(ellipse_array)
        s1.add_ellipse_array_as_markers(
            ellipse_array, inlier_array=inlier_array.T, peak_array=parray.T
        )

    def test_wrong_input_dimensions(self):
        s = Diffraction2D(np.zeros((2, 5, 5)))
        with pytest.raises(ValueError):
            s.add_ellipse_array_as_markers(ellipse_array=None)


class TestDiffraction2DThresholdAndMask:
    @pytest.mark.parametrize("shape", [(9, 9, 9, 9), (4, 8, 6, 3), (6, 3, 2, 5)])
    def test_no_change(self, shape):
        s = Diffraction2D(np.zeros(shape))
        s1 = s.threshold_and_mask()
        assert (s.data == s1.data).all()

    @pytest.mark.parametrize("mask", [(3, 5, 1), (6, 8, 1), (4, 6, 1)])
    def test_mask(self, mask):
        s = Diffraction2D(np.ones((10, 10, 10, 10)))
        s1 = s.threshold_and_mask(mask=mask)
        slice0 = np.s_[:, :, mask[1] - mask[2] : mask[1] + mask[2] + 1, mask[0]]
        assert (s1.data[slice0] == 1.0).all()
        slice1 = np.s_[:, :, mask[1], mask[0] - mask[2] : mask[0] + mask[2] + 1]
        assert (s1.data[slice1] == 1.0).all()
        s1.data[slice0] = 0
        s1.data[slice1] = 0
        assert (s1.data == 0).all()

    @pytest.mark.parametrize("x, y", [(3, 5), (7, 5), (5, 2)])
    def test_threshold(self, x, y):
        s = Diffraction2D(np.random.randint(0, 10, size=(10, 10, 10, 10)))
        s.data[:, :, x, y] = 1000000
        s1 = s.threshold_and_mask(threshold=1)
        assert (s1.data[:, :, x, y] == 1.0).all()
        s1.data[:, :, x, y] = 0
        assert (s1.data == 0).all()

    def test_threshold_mask(self):
        s = Diffraction2D(np.zeros((12, 11, 13, 10)))
        s.data[:, :, 1, 2] = 1000000
        s.data[:, :, 8, 6] = 10
        s1 = s.threshold_and_mask(threshold=1)
        assert (s1.data[:, :, 1, 2] == 1.0).all()
        s1.data[:, :, 1, 2] = 0
        assert (s1.data == 0).all()

        s2 = s.threshold_and_mask(threshold=1, mask=(6, 8, 1))
        assert (s2.data[:, :, 8, 6] == 1.0).all()
        s2.data[:, :, 8, 6] = 0
        assert (s2.data == 0).all()

    def test_lazy_exception(self):
        s = dd.get_disk_shift_simple_test_signal(lazy=True)
        with pytest.raises(NotImplementedError):
            s.threshold_and_mask()


class TestDiffraction2DAngleSector:
    def test_get_angle_sector_mask_simple(self):
        array = np.zeros((10, 10, 10, 10))
        array[:, :, 0:5, 0:5] = 1
        s = Diffraction2D(array)
        s.axes_manager.signal_axes[0].offset = -4.5
        s.axes_manager.signal_axes[1].offset = -4.5
        mask = s.angular_mask(0.0, 0.5 * np.pi)
        assert mask[:, :, 0:5, 0:5].all()
        assert not mask[:, :, 5:, :].any()
        assert not mask[:, :, :, 5:].any()

    def test_com_angle_sector_mask(self):
        x, y = 4, 7
        array = np.zeros((5, 4, 10, 20))
        array[:, :, y, x] = 1
        s = Diffraction2D(array)
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        s.angular_mask(
            0.0,
            0.5 * np.pi,
            centre_x_array=s_bs.isig[0].data,
            centre_y_array=s_bs.isig[1].data,
        )


class TestAngularSliceRadialAverage:
    def test_deprecated_method(self):
        s = Diffraction2D(np.zeros((2, 2, 10, 10)))
        with pytest.raises(Exception):
            s.angular_slice_radial_integration()


class TestDiffraction2DTemplateMatchDisk:
    def test_simple(self):
        s = Diffraction2D(np.random.randint(100, size=(5, 5, 20, 20)))
        s_template = s.template_match_disk()
        assert s.data.shape == s_template.data.shape
        assert not s_template._lazy

    def test_disk_radius(self):
        s = LazyDiffraction2D(
            da.random.randint(100, size=(5, 5, 30, 30), chunks=(1, 1, 10, 10))
        )
        s_template0 = s.template_match_disk(disk_r=2, lazy_output=False)
        s_template1 = s.template_match_disk(disk_r=4, lazy_output=False)
        assert s.data.shape == s_template0.data.shape
        assert s.data.shape == s_template1.data.shape
        assert not (s_template0.data == s_template1.data).all()

    @pytest.mark.parametrize("mode", ["constant", "reflect"])
    def test_template_circular(self, mode):
        s = Diffraction2D(np.random.randint(100, size=(5, 5, 20, 20)))
        s_template = s.template_match_disk(circular_background=True, mode=mode)
        assert s.data.shape == s_template.data.shape
        assert not s_template._lazy


class TestDiffraction2DTemplateMatchRing:
    def test_simple(self):
        s = Diffraction2D(np.random.randint(100, size=(5, 5, 20, 20)))
        s_template = s.template_match_ring(r_inner=3, r_outer=5, lazy_output=True)
        assert s.data.shape == s_template.data.shape
        assert s_template._lazy is True

    def test_wrong_input(self):
        s = Diffraction2D(np.random.randint(100, size=(5, 5, 20, 20)))
        with pytest.raises(ValueError):
            s.template_match_ring(r_inner=5, r_outer=3)
        with pytest.raises(ValueError):
            s.template_match_ring(r_inner=3, r_outer=3)


class TestDiffraction2DTemplate:
    def test_square_and_disk(self):
        s = Diffraction2D(np.zeros((2, 2, 100, 100)))

        square_image = np.zeros((9, 9))
        square_image[2:-2, 2:-2] = 1
        s.data[:, :, 20:29, 40:49] = square_image

        disk = morphology.disk(4, s.data.dtype)
        s.data[:, :, 60:69, 50:59] = disk

        s_st = s.template_match(square_image)
        s_dt = s.template_match(disk)

        st_ind = np.unravel_index(np.argmax(s_st.data, axis=None), s_st.data.shape)[-2:]
        dt_ind = np.unravel_index(np.argmax(s_dt.data, axis=None), s_dt.data.shape)[-2:]

        assert st_ind == (24, 44)
        assert dt_ind == (64, 54)
        assert s.data.shape == s_st.data.shape
        assert s.data.shape == s_dt.data.shape


# Remove in 1.0.0 Release
class TestDiffraction2DTemplateWithBinaryImage:
    def test_square_and_disk(self):
        s = Diffraction2D(np.zeros((2, 2, 100, 100)))

        square_image = np.zeros((9, 9))
        square_image[2:-2, 2:-2] = 1
        s.data[:, :, 20:29, 40:49] = square_image

        disk = morphology.disk(4, s.data.dtype)
        s.data[:, :, 60:69, 50:59] = disk

        s_st = s.template_match_with_binary_image(square_image, lazy_result=False)
        s_dt = s.template_match_with_binary_image(disk, lazy_result=False)

        st_ind = np.unravel_index(np.argmax(s_st.data, axis=None), s_st.data.shape)[-2:]
        dt_ind = np.unravel_index(np.argmax(s_dt.data, axis=None), s_dt.data.shape)[-2:]

        assert st_ind == (24, 44)
        assert dt_ind == (64, 54)
        assert s.data.shape == s_st.data.shape
        assert s.data.shape == s_dt.data.shape


class TestDiffraction2DRotateDiffraction:
    def test_rotate_diffraction_keep_shape(self):
        shape = (7, 5, 4, 15)
        s = Diffraction2D(np.zeros(shape))
        s_rot = s.rotate_diffraction(angle=45)
        assert s.axes_manager.shape == s_rot.axes_manager.shape

        s_lazy = LazyDiffraction2D(da.zeros(shape, chunks=(1, 1, 1, 1)))
        s_rot_lazy = s_lazy.rotate_diffraction(angle=45)
        assert s_lazy.axes_manager.shape == s_rot_lazy.axes_manager.shape

    def test_rotate_diffraction_values(self):
        data = np.zeros((10, 5, 12, 14))
        data[:, :, 6:, 7:] = 1
        s = Diffraction2D(data)
        s_rot = s.rotate_diffraction(angle=180)
        np.testing.assert_almost_equal(
            s.data[0, 0, :6, :7], np.zeros_like(s.data[0, 0, :6, :7])
        )
        np.testing.assert_almost_equal(
            s_rot.data[0, 0, :6, :7], np.ones_like(s_rot.data[0, 0, :6, :7])
        )
        s_rot.data[:, :, :6, :7] = 0
        np.testing.assert_almost_equal(s_rot.data, np.zeros_like(s.data))


class TestDiffraction2DShiftDiffraction:
    @pytest.mark.parametrize("shift_x,shift_y", [(2, 5), (-6, -1), (2, -4)])
    def test_single_shift(self, shift_x, shift_y):
        s = Diffraction2D(np.zeros((10, 10, 30, 40)))
        x, y = 20, 10
        s.data[:, :, y, x] = 1
        s_shift = s.shift_diffraction(shift_x=shift_x, shift_y=shift_y)
        assert s_shift.data[0, 0, y - shift_y, x - shift_x] == 1
        s_shift.data[:, :, y - shift_y, x - shift_x] = 0
        assert s_shift.data.sum() == 0

    @pytest.mark.parametrize("centre_x,centre_y", [(25, 25), (30, 20)])
    def test_random_shifts(self, centre_x, centre_y):
        y, x = np.mgrid[20:30:7j, 20:30:5j]
        s = mdtd.generate_4d_data(
            probe_size_x=5,
            probe_size_y=7,
            disk_x=x,
            disk_y=y,
            disk_r=1,
            blur=True,
            ring_x=None,
        )
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        s_shift = s.shift_diffraction(
            shift_x=-s_bs.isig[0].data, shift_y=-s_bs.isig[1].data
        )
        s_shift_c = s_shift.get_direct_beam_position(method="center_of_mass")
        assert s_shift_c.data == pytest.approx(0.0)

    def test_inplace(self):
        s = Diffraction2D(np.zeros((10, 10, 30, 40)))
        x, y, shift_x, shift_y = 20, 10, 4, -3
        s.data[:, :, y, x] = 1
        s.shift_diffraction(shift_x=shift_x, shift_y=shift_y, inplace=True)
        assert s.data[0, 0, y - shift_y, x - shift_x] == 1
        s.data[:, :, y - shift_y, x - shift_x] = 0
        assert s.data.sum() == 0

    def test_lazy(self):
        data = np.zeros((10, 10, 30, 30))
        x, y = 20, 10
        data[:, :, y, x] += 1
        data = da.from_array(data, chunks=(5, 5, 5, 5))
        s = LazyDiffraction2D(data)
        shift_x, shift_y = 4, 3
        s_shift = s.shift_diffraction(shift_x=shift_x, shift_y=shift_y)
        s_shift.compute()
        assert s_shift.data[0, 0, y - shift_y, x - shift_x] == 1
        s_shift.data[:, :, y - shift_y, x - shift_x] = 0
        assert s_shift.data.sum() == 0


class TestComputeAndAsLazy:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyDiffraction2D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert s.__class__ == Diffraction2D
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_5d_data_compute(self):
        dask_array = da.random.random((2, 3, 4, 10, 15), chunks=(1, 1, 1, 10, 15))
        s = LazyDiffraction2D(dask_array)
        s.compute()
        assert s.__class__ == Diffraction2D
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = Diffraction2D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyDiffraction2D
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_5d_data_as_lazy(self):
        data = np.random.random((2, 3, 4, 10, 15))
        s = Diffraction2D(data)
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyDiffraction2D
        assert data.shape == s_lazy.data.shape



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.signals import ReducedIntensity1D
from pyxem.components import ScatteringFitComponentLobato


def test_scattering_component_init_lobato():
    ref = ScatteringFitComponentLobato(["Cu"], [1], N=1.0, C=0.0)
    assert isinstance(ref, ScatteringFitComponentLobato)


@pytest.fixture(params=[np.array([4.0, 3.0, 2.0, 2.0, 1.0, 1.0, 1.0, 0.0])])
def ri_model(request):
    ri = ReducedIntensity1D(request.param)
    m = ri.create_model()
    return m


def test_function_lobato(ri_model):
    elements = ["Cu"]
    fracs = [1]
    sc_component = ScatteringFitComponentLobato(elements, fracs, N=1.0, C=0.0)
    ri_model.append(sc_component)
    ri_model.fit()



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from sklearn.cluster import DBSCAN
from skimage.draw import disk
from skimage.morphology import disk as disk2

from hyperspy.signals import Signal2D
from hyperspy.signal import BaseSignal


from pyxem.signals import DiffractionVectors, DiffractionVectors2D
from pyxem.utils._subpixel_finding import (
    _center_of_mass_hs,
    _get_experimental_square,
    _conventional_xc,
)
from pyxem.data import tilt_boundary_data
from hyperspy.axes import UniformDataAxis

# DiffractionVectors correspond to a single list of vectors, a map of vectors
# all of equal length, and the ragged case. A fixture is defined for each of
# these cases and all methods tested for it.


@pytest.fixture(
    params=[
        np.array(
            [
                [
                    np.array(
                        [
                            [0.089685, 0.292971],
                            [0.017937, 0.277027],
                            [-0.069755, 0.257097],
                            [-0.165419, 0.241153],
                            [0.049825, 0.149475],
                            [-0.037867, 0.129545],
                            [-0.117587, 0.113601],
                        ]
                    ),
                    np.array(
                        [
                            [0.089685, 0.292971],
                            [0.017937, 0.277027],
                            [-0.069755, 0.257097],
                            [-0.165419, 0.241153],
                            [0.049825, 0.149475],
                            [-0.037867, 0.129545],
                            [-0.117587, 0.113601],
                            [0.149475, 0.065769],
                            [0.229195, 0.045839],
                            [0.141503, 0.025909],
                            [0.073741, 0.013951],
                            [0.001993, 0.001993],
                            [-0.069755, -0.009965],
                        ]
                    ),
                ],
                [
                    np.array(
                        [
                            [0.089685, 0.292971],
                            [0.017937, 0.277027],
                            [-0.069755, 0.257097],
                            [-0.165419, 0.241153],
                            [0.049825, 0.149475],
                            [-0.037867, 0.129545],
                            [-0.117587, 0.113601],
                            [0.149475, 0.065769],
                            [0.229195, 0.045839],
                            [0.141503, 0.025909],
                            [0.073741, 0.013951],
                        ]
                    ),
                    np.array([[0.001993, 0.001993]]),
                ],
            ],
            dtype=object,
        )
    ]
)
def diffraction_vectors_map(request):
    dvm = DiffractionVectors(request.param)
    dvm.axes_manager[0].name = "x"
    dvm.axes_manager[1].name = "y"
    return dvm


class TestVectorPlotting:
    def test_plot_diffraction_vectors(self, diffraction_vectors_map):
        with pytest.warns(UserWarning, match="distance_threshold=0 was given"):
            diffraction_vectors_map.plot_diffraction_vectors(
                xlim=1.0, ylim=1.0, distance_threshold=0
            )

    def test_plot_diffraction_vectors_on_signal(
        self, diffraction_vectors_map, diffraction_pattern
    ):
        diffraction_vectors_map.plot_diffraction_vectors_on_signal(diffraction_pattern)


def test_get_cartesian_coordinates(diffraction_vectors_map):
    accelerating_voltage = 200
    camera_length = 0.2
    diffraction_vectors_map.calculate_cartesian_coordinates(
        accelerating_voltage, camera_length
    )
    # Coordinate conversion is tested in vector_utils. Just test that the
    # result is stored correctly
    assert diffraction_vectors_map.cartesian is not None
    assert (
        diffraction_vectors_map.axes_manager[0].name
        == diffraction_vectors_map.cartesian.axes_manager[0].name
    )


class TestInitVectors:
    @pytest.fixture()
    def peaks(self):
        vectors = np.empty((2, 2), dtype=object)
        vectors[0, 0] = np.random.randint(0, 100, (5, 2))
        vectors[0, 1] = np.random.randint(0, 100, (6, 2))
        vectors[1, 0] = np.random.randint(0, 100, (7, 2))
        vectors[1, 1] = np.random.randint(0, 100, (8, 2))

        peaks = BaseSignal(vectors, ragged=True)
        return peaks

    @pytest.fixture()
    def peaks_w_intensity(self):
        vectors = np.empty((2, 2), dtype=object)
        vectors[0, 0] = np.random.randint(0, 100, (5, 3))
        vectors[0, 1] = np.random.randint(0, 100, (6, 3))
        vectors[1, 0] = np.random.randint(0, 100, (7, 3))
        vectors[1, 1] = np.random.randint(0, 100, (8, 3))

        peaks = BaseSignal(vectors, ragged=True)
        return peaks

    @pytest.mark.parametrize("column_names", (None, ["x", "y"]))
    @pytest.mark.parametrize("units", (None, ["nm", "nm"]))
    def test_from_peaks(self, peaks, column_names, units):
        peaks.axes_manager.navigation_axes[0].name = "x"
        peaks.axes_manager.navigation_axes[1].name = "y"
        dv = DiffractionVectors.from_peaks(
            peaks,
            center=(50, 50),
            calibration=0.1,
            column_names=column_names,
            units=units,
        )

        for i in np.ndindex((2, 2)):
            np.testing.assert_array_equal((peaks.data[i] - 50) * 0.1, dv.data[i])
        assert dv.scales == [0.1, 0.1]
        assert dv.axes_manager[0].name == "x"
        assert dv.axes_manager[1].name == "y"
        assert dv.column_names == ["x", "y"]
        if units is not None:
            assert dv.units == ["nm", "nm"]

    def test_from_peaks_lazy(self, peaks):
        peaks = peaks.as_lazy()
        dv = DiffractionVectors.from_peaks(
            peaks,
            center=(50, 50),
            calibration=0.1,
        )
        assert dv._lazy

        for i in np.ndindex((2, 2)):
            np.testing.assert_array_equal((peaks.data[i] - 50) * 0.1, dv.data[i])
        assert dv.scales == [0.1, 0.1]

    def test_from_peaks_calibration(self, peaks):
        peaks.metadata.add_node("Peaks.signal_axes")
        peaks.metadata.Peaks.signal_axes = (
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
        )
        dv = DiffractionVectors.from_peaks(
            peaks,
            center=None,
            calibration=None,
        )

        for i in np.ndindex((2, 2)):
            np.testing.assert_almost_equal((peaks.data[i]) * 0.1 - 5.0, dv.data[i])
        assert dv.scales == [0.1, 0.1]

    def test_from_peaks_calibration_to_markers(self, peaks):
        peaks.metadata.add_node("Peaks.signal_axes")
        peaks.metadata.Peaks.signal_axes = (
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
        )
        dv = DiffractionVectors.from_peaks(
            peaks,
            center=None,
            calibration=None,
        )
        points = dv.to_markers()
        for i in np.ndindex((2, 2)):
            np.testing.assert_almost_equal(
                ((peaks.data[i]) * 0.1 - 5.0)[:, 0],
                points.kwargs["offsets"][i[::-1]][:, 1],
            )

    def test_from_peaks_calibration_to_markers_with_intensity(self, peaks_w_intensity):
        peaks_w_intensity.metadata.add_node("Peaks.signal_axes")
        peaks_w_intensity.metadata.Peaks.signal_axes = (
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
        )
        dv = DiffractionVectors.from_peaks(
            peaks_w_intensity,
            center=None,
            calibration=None,
        )
        points = dv.to_markers()
        for i in np.ndindex((2, 2)):
            np.testing.assert_almost_equal(
                ((peaks_w_intensity.data[i]) * 0.1 - 5.0)[:, 0],
                points.kwargs["offsets"][i[::-1]][:, 1],
            )
        s = Signal2D(np.ones((2, 2, 10, 10)))
        s.add_marker(points)

    def test_from_peaks_calibration_error(self, peaks):
        with pytest.raises(ValueError):
            dv = DiffractionVectors.from_peaks(
                peaks,
                center=None,
                calibration=None,
            )
        with pytest.raises(ValueError):
            dv = DiffractionVectors.from_peaks(
                peaks,
                center=(1.0, 1.0),
                calibration=None,
            )
        with pytest.raises(ValueError):
            dv = DiffractionVectors.from_peaks(
                peaks,
                center=None,
                calibration=(1.0, 1.0),
            )

    def test_from_peaks_calibration_2_peaks(self, peaks):
        peaks.metadata.add_node("Peaks.signal_axes")
        peaks.metadata.Peaks.signal_axes = (
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
        )
        dv = DiffractionVectors.from_peaks(
            peaks,
            center=None,
            calibration=None,
        )
        pixels = dv.pixel_vectors

        for i in np.ndindex((2, 2)):
            np.testing.assert_almost_equal(peaks.data[i], pixels.data[i])

    def test_peaks_with_intensity(self, peaks_w_intensity):
        peaks_w_intensity.metadata.add_node("Peaks.signal_axes")
        peaks_w_intensity.metadata.Peaks.signal_axes = (
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
            UniformDataAxis(scale=0.1, offset=-5.0, units="nm"),
        )
        dv = DiffractionVectors.from_peaks(
            peaks_w_intensity,
            center=None,
            calibration=None,
        )
        pixels = dv.pixel_vectors
        for i in np.ndindex((2, 2)):
            np.testing.assert_almost_equal(peaks_w_intensity.data[i], pixels.data[i])

    def test_initial_metadata(self, diffraction_vectors_map):
        assert diffraction_vectors_map.scales is None
        assert diffraction_vectors_map.metadata.VectorMetadata["scales"] == None

        assert diffraction_vectors_map.offsets is None
        assert diffraction_vectors_map.metadata.VectorMetadata["offsets"] == None

        assert diffraction_vectors_map.detector_shape is None
        assert diffraction_vectors_map.metadata.VectorMetadata["detector_shape"] == None

    def test_set_scales(self, diffraction_vectors_map):
        diffraction_vectors_map.scales = 0.1
        assert diffraction_vectors_map.scales == [0.1, 0.1]
        assert (
            diffraction_vectors_map.scales == diffraction_vectors_map.pixel_calibration
        )

    def test_scales_error(self, diffraction_vectors_map):
        with pytest.raises(ValueError):
            diffraction_vectors_map.scales = [1, 2, 3]

    def test_set_column_names_error(self, diffraction_vectors_map):
        with pytest.raises(ValueError):
            diffraction_vectors_map.column_names = ["x", "y", "z"]

    def test_set_units_error(self, diffraction_vectors_map):
        with pytest.raises(ValueError):
            diffraction_vectors_map.units = ["x", "y", "z"]

    def test_set_units_singleton(self, diffraction_vectors_map):
        diffraction_vectors_map.units = "nm"
        assert diffraction_vectors_map.units == ["nm", "nm"]

    def test_set_column_names(self, diffraction_vectors_map):
        diffraction_vectors_map.column_names = ["x", "y"]
        assert diffraction_vectors_map.column_names == ["x", "y"]

    def test_set_column_names_none(self, diffraction_vectors_map):
        diffraction_vectors_map.column_names = None
        assert diffraction_vectors_map.column_names == ["column_0", "column_1"]

    def test_num_rows(self, diffraction_vectors_map):
        assert diffraction_vectors_map.num_rows is None

    def test_set_offsets_error(self, diffraction_vectors_map):
        with pytest.raises(ValueError):
            diffraction_vectors_map.offsets = [1, 2, 3]

    def test_setting_metadat(self, diffraction_vectors_map):
        diffraction_vectors_map.scales = 0.1
        assert diffraction_vectors_map.metadata.VectorMetadata["scales"] == [0.1, 0.1]

        diffraction_vectors_map.offsets = 1
        assert diffraction_vectors_map.metadata.VectorMetadata["offsets"] == [1, 1]

        diffraction_vectors_map.detector_shape = [100, 100]
        assert diffraction_vectors_map.metadata.VectorMetadata["detector_shape"] == [
            100,
            100,
        ]


class TestSubpixelRefinement:
    def test_com_accuracy(self):
        xx, yy = disk(center=(45, 50), radius=10)
        cir = np.zeros((100, 100))
        cir[xx, yy] = 1
        sq = _get_experimental_square(cir, (47, 50), 30)
        cen = _center_of_mass_hs(sq)
        assert np.allclose(np.array(cen) - 15, (-2.0, 0.0))

    def test_conventional_xc_accuracy(self):
        xx, yy = disk(center=(45, 50), radius=10)
        kernel = disk2(radius=10)
        cir = np.random.random((100, 100))

        cir[xx, yy] += 10
        sq = _get_experimental_square(cir, (47, 50), 30)
        cen = _conventional_xc(sq, kernel, upsample_factor=1)
        assert np.allclose(np.array(cen), (-2.0, 0.0))

    @pytest.mark.parametrize("intensity", (True, False))
    def test_subpixel_refinement_com(self, intensity):
        import pyxem.data.dummy_data.make_diffraction_test_data as mdtd
        import pyxem as pxm
        import numpy as np

        di = mdtd.DiffractionTestImage(intensity_noise=False)
        di.add_disk(x=128, y=128, intensity=10.0)  # Add a zero beam disk at the center
        di.add_cubic_disks(vx=20, vy=20, intensity=2.0, n=5)
        di.add_background_lorentz()
        dtd = mdtd.DiffractionTestDataset(10, 10, 256, 256)
        position_array = np.ones((10, 10), dtype=bool)
        position_array[:5] = False
        dtd.add_diffraction_image(di)
        s = dtd.get_signal()
        temp = s.template_match_disk(disk_r=5, subtract_min=False)
        dv = temp.get_diffraction_vectors(
            threshold_abs=0.4, min_distance=5, get_intensity=intensity
        )
        dv.subpixel_refine(s, method="center-of-mass")
        if intensity:
            assert dv.data[0, 0].shape[1] == 3
        else:
            assert dv.data[0, 0].shape[1] == 2

    @pytest.mark.parametrize("intensity", (True, False))
    def test_subpixel_refinement_xc(self, intensity):
        import pyxem.data.dummy_data.make_diffraction_test_data as mdtd
        import pyxem as pxm
        import numpy as np

        di = mdtd.DiffractionTestImage(intensity_noise=False)
        di.add_disk(x=128, y=128, intensity=10.0)  # Add a zero beam disk at the center
        di.add_cubic_disks(vx=20, vy=20, intensity=2.0, n=5)
        di.add_background_lorentz()
        dtd = mdtd.DiffractionTestDataset(10, 10, 256, 256)
        position_array = np.ones((10, 10), dtype=bool)
        position_array[:5] = False
        dtd.add_diffraction_image(di)
        s = dtd.get_signal()
        temp = s.template_match_disk(disk_r=5, subtract_min=False)
        dv = temp.get_diffraction_vectors(
            threshold_abs=0.4, min_distance=5, get_intensity=intensity
        )
        dv.subpixel_refine(s, method="cross-correlation", upsample_factor=2, disk_r=5)
        if intensity:
            assert dv.data[0, 0].shape[1] == 3
        else:
            assert dv.data[0, 0].shape[1] == 2


class TestConvertVectors:
    @pytest.mark.parametrize("real_units", (True, False))
    def test_flatten_vectors(self, diffraction_vectors_map, real_units):
        vectors = diffraction_vectors_map.flatten_diffraction_vectors(
            real_units=real_units
        )
        assert isinstance(vectors, DiffractionVectors2D)
        assert vectors.data.shape == (32, 4)
        assert vectors.ragged == False

    def test_flatten_vectors1d(self, diffraction_vectors_map):
        oned_vectors = diffraction_vectors_map.ivec[0]
        vectors = oned_vectors.flatten_diffraction_vectors()

    def test_get_navigation_positions(self):
        test_data = np.empty((2, 3), dtype=object)
        for i in np.ndindex(test_data.shape):
            test_data[i] = np.random.random((np.random.randint(2, 6), 2))
        dv = DiffractionVectors(test_data)
        dv.axes_manager.navigation_axes.set(
            name=["x", "y"], scale=[1, 2], offset=[1, 0]
        )
        flat = dv._get_navigation_positions(flatten=False, real_units=True)
        np.testing.assert_array_equal(flat[0, 0, 0], 0)
        np.testing.assert_array_equal(flat[0, 0, 1], 1)
        np.testing.assert_array_equal(flat[0, 1, 1], 2)
        np.testing.assert_array_equal(flat[0, 1, 0], 0)

    def test_flatten_vectors_with_scales(self):
        test_data = np.empty((2, 3, 4), dtype=object)
        for i in np.ndindex(test_data.shape):
            test_data[i] = np.random.random((np.random.randint(2, 6), 2))
        dv = DiffractionVectors(test_data)
        dv.axes_manager.navigation_axes.set(
            name=["x", "y", "time"], scale=[1, 2, 3], offset=[1, 2, 3]
        )
        flat = dv.flatten_diffraction_vectors(real_units=True)
        assert flat.column_names[0] == "x"
        assert flat.column_names[1] == "y"
        assert flat.column_names[2] == "time"
        assert (
            np.max(flat.data[:, 0])
            == dv.axes_manager[0].size * dv.axes_manager[0].scale
        )
        assert (
            np.max(flat.data[:, 1])
            == dv.axes_manager[1].size * dv.axes_manager[1].scale
        )
        assert (
            np.max(flat.data[:, 2])
            == dv.axes_manager[2].size * dv.axes_manager[2].scale
        )
        assert np.min(flat.data[:, 0]) == dv.axes_manager[0].offset
        assert np.min(flat.data[:, 1]) == dv.axes_manager[1].offset
        assert np.min(flat.data[:, 2]) == dv.axes_manager[2].offset

    def test_flatten_vectors_with_set_metadata(self, diffraction_vectors_map):
        diffraction_vectors_map.scales = [0.1, 0.1]
        diffraction_vectors_map.offsets = [1, 1]
        vectors = diffraction_vectors_map.flatten_diffraction_vectors(real_units=True)

        assert isinstance(vectors, DiffractionVectors2D)
        assert vectors.data.shape == (32, 4)


class TestMagnitudes:
    def test_get_magnitudes_map(self, diffraction_vectors_map):
        diffraction_vectors_map.get_magnitudes()

    @pytest.mark.filterwarnings("ignore::FutureWarning")  # deemed "safe enough"
    def test_get_magnitude_histogram_map(self, diffraction_vectors_map):
        diffraction_vectors_map.get_magnitude_histogram(bins=np.arange(0, 0.5, 0.1))


class TestUniqueVectors:
    def test_get_unique_vectors_map_type(self, diffraction_vectors_map):
        unique_vectors = diffraction_vectors_map.get_unique_vectors()
        assert isinstance(unique_vectors, DiffractionVectors2D)

    @pytest.mark.parametrize(
        "distance_threshold, answer",
        [
            (
                0.01,
                np.array(
                    [
                        [-0.165419, 0.241153],
                        [-0.117587, 0.113601],
                        [-0.069755, -0.009965],
                        [-0.069755, 0.257097],
                        [-0.037867, 0.129545],
                        [0.001993, 0.001993],
                        [0.017937, 0.277027],
                        [0.049825, 0.149475],
                        [0.073741, 0.013951],
                        [0.089685, 0.292971],
                        [0.141503, 0.025909],
                        [0.149475, 0.065769],
                        [0.229195, 0.045839],
                    ]
                ),
            ),
            (
                0.1,
                np.array(
                    [
                        [-0.117587, 0.249125],
                        [-0.077727, 0.121573],
                        [-0.021923, -0.001993],
                        [0.053811, 0.284999],
                        [0.049825, 0.149475],
                        [0.121573, 0.03520967],
                        [0.229195, 0.045839],
                    ]
                ),
            ),
        ],
    )
    def test_get_unique_vectors_map_values(
        self, diffraction_vectors_map, distance_threshold, answer
    ):
        unique_vectors = diffraction_vectors_map.get_unique_vectors(
            distance_threshold=distance_threshold
        )
        np.testing.assert_almost_equal(unique_vectors.data, answer)

    def test_get_unique_vectors_map_dbscan(self, diffraction_vectors_map):
        unique_dbscan = diffraction_vectors_map.get_unique_vectors(
            method="DBSCAN", return_clusters=True
        )
        assert isinstance(unique_dbscan[0], DiffractionVectors2D)
        assert isinstance(unique_dbscan[1], DBSCAN)

    @pytest.mark.parametrize(
        "distance_threshold, answer",
        [
            (
                0.01,
                np.array(
                    [
                        [-0.165419, 0.241153],
                        [-0.117587, 0.113601],
                        [-0.069755, -0.009965],
                        [-0.069755, 0.257097],
                        [-0.037867, 0.129545],
                        [0.001993, 0.001993],
                        [0.017937, 0.277027],
                        [0.049825, 0.149475],
                        [0.073741, 0.013951],
                        [0.089685, 0.292971],
                        [0.141503, 0.025909],
                        [0.149475, 0.065769],
                        [0.229195, 0.045839],
                    ]
                ),
            ),
            (
                0.1,
                np.array(
                    [
                        [-0.031888, 0.267062],
                        [-0.03520967, 0.13087367],
                        [0.10200536, 0.02699609],
                    ]
                ),
            ),
        ],
    )
    def test_get_unique_vectors_map_values_dbscan(
        self, diffraction_vectors_map, distance_threshold, answer
    ):
        unique_vectors = diffraction_vectors_map.get_unique_vectors(
            distance_threshold=distance_threshold, method="DBSCAN"
        )
        np.testing.assert_almost_equal(unique_vectors.data, answer)


class TestFilterVectors:
    def test_filter_magnitude_map_type(self, diffraction_vectors_map):
        filtered_vectors = diffraction_vectors_map.filter_magnitude(0.1, 1.0)
        assert isinstance(filtered_vectors, DiffractionVectors)

    def test_filter_magnitude_map(self, diffraction_vectors_map):
        filtered_vectors = diffraction_vectors_map.filter_magnitude(0.1, 1.0)
        ans = np.array(
            [
                [0.089685, 0.292971],
                [0.017937, 0.277027],
                [-0.069755, 0.257097],
                [-0.165419, 0.241153],
                [0.049825, 0.149475],
                [-0.037867, 0.129545],
                [-0.117587, 0.113601],
                [0.149475, 0.065769],
                [0.229195, 0.045839],
                [0.141503, 0.025909],
            ]
        )
        np.testing.assert_almost_equal(filtered_vectors.data[0][1], ans)

    def test_filter_detector_edge_map_type(self, diffraction_vectors_map):
        diffraction_vectors_map.detector_shape = (260, 240)
        diffraction_vectors_map.pixel_calibration = 0.001
        filtered_vectors = diffraction_vectors_map.filter_detector_edge(exclude_width=2)
        assert isinstance(filtered_vectors, DiffractionVectors)

    def test_filter_detector_edge_map(self, diffraction_vectors_map):
        vectors = diffraction_vectors_map.deepcopy()
        vectors.detector_shape = (260, 240)
        vectors.pixel_calibration = 0.001

        filtered_vectors = vectors.filter_detector_edge(exclude_width=2)
        ans = np.array([[-0.117587, 0.113601]])
        np.testing.assert_almost_equal(filtered_vectors.data[0, 0], ans)

    def test_filter_basis(self):
        basis = np.array(
            [
                [0.089685, 0.292971],
                [0.017937, 0.277027],
                [-0.069755, 0.257097],
                [-0.165419, 0.241153],
                [0.049825, 0.149475],
                [-0.037867, 0.129545],
                [-0.117587, 0.113601],
            ]
        )
        arr = np.empty(2, dtype=object)
        arr[0] = basis
        arr[1] = basis[:-1]
        vect = DiffractionVectors(arr)
        filtered = vect.filter_basis(basis=basis, distance=0.1)
        assert isinstance(filtered, DiffractionVectors2D)
        np.testing.assert_almost_equal(filtered.data[0], basis)

    def test_filter_basis_ragged(self, diffraction_vectors_map):
        basis = diffraction_vectors_map.deepcopy()
        filtered = diffraction_vectors_map.filter_basis(basis=basis, distance=0.1)
        assert isinstance(filtered, DiffractionVectors)
        np.testing.assert_almost_equal(
            filtered.data[0, 0], diffraction_vectors_map.data[0, 0]
        )


class TestDiffractingPixelsMap:
    def test_get_dpm_values(self, diffraction_vectors_map):
        answer = np.array([[7.0, 13.0], [11.0, 1.0]])
        xim = diffraction_vectors_map.get_diffracting_pixels_map()
        assert np.allclose(xim, answer)

    def test_get_dpm_type(self, diffraction_vectors_map):
        xim = diffraction_vectors_map.get_diffracting_pixels_map()
        assert isinstance(xim, Signal2D)

    def test_get_dpm_title(self, diffraction_vectors_map):
        xim = diffraction_vectors_map.get_diffracting_pixels_map()
        assert xim.metadata.General.title == "Diffracting Pixels Map"

    def test_get_dpm_in_range(self, diffraction_vectors_map):
        answer = np.array([[0.0, 3.0], [1.0, 1.0]])
        xim = diffraction_vectors_map.get_diffracting_pixels_map(in_range=(0, 0.1))
        assert np.allclose(xim, answer)

    def test_get_dpm_binary(self, diffraction_vectors_map):
        answer = np.array([[1.0, 1.0], [1.0, 1.0]])
        xim = diffraction_vectors_map.get_diffracting_pixels_map(binary=True)
        assert np.allclose(xim, answer)


class TestSlicingVectors:
    @pytest.fixture()
    def vectors(self):
        vectors = np.empty((2, 2), dtype=object)
        vectors[0, 0] = np.random.randint(-100, 100, (20, 2))
        vectors[0, 1] = np.random.randint(-100, 100, (6, 2))
        vectors[1, 0] = np.random.randint(-100, 100, (7, 2))
        vectors[1, 1] = np.random.randint(-100, 100, (1, 2))
        v = DiffractionVectors(
            vectors,
            scales=[0.1, 0.2],
            offsets=[10, 20],
            column_names=["x_axis", "y_axis"],
        )

        return v

    def test_repr(self, vectors):
        repr = vectors._repr_html_()
        assert isinstance(repr, str)

    def test_center(self, vectors):
        np.testing.assert_almost_equal(vectors.center, (100, 100))

    @pytest.mark.parametrize("index", (0, "x_axis", ("x_axis",)))
    def test_column(self, vectors, index):
        slic = vectors.ivec[index]
        for i in np.ndindex((2, 2)):
            np.testing.assert_almost_equal(slic.data[i][:], vectors.data[i][:, 0])

    def test_column_error(self, vectors):
        with pytest.raises(ValueError):
            vectors.ivec[5.5]

    @pytest.mark.parametrize("index", ([0, 1], ["x_axis", "y_axis"]))
    def test_column_slicing2(self, vectors, index):
        slic = vectors.ivec[index]
        for i in np.ndindex((2, 2)):
            np.testing.assert_almost_equal(slic.data[i][:, [0, 1]], vectors.data[i])

    @pytest.mark.parametrize("item", [0, "x_axis"])
    def test_slice(self, vectors, item):
        sliced = vectors.ivec[item]
        assert sliced.column_names == [
            "x_axis",
        ]

    def test_row_lt(self, vectors):
        col = vectors.ivec[0] < 0.5
        slic = vectors.ivec[:, col]
        for i in np.ndindex((2, 2)):
            assert np.all(slic.data[i][:, 0] < 0.5)

    def test_row_gt(self, vectors):
        col = vectors.ivec[0] > 0.5
        slic = vectors.ivec[:, col]
        for i in np.ndindex((2, 2)):
            assert np.all(slic.data[i][:, 0] > 0.5)

    def test_row_gt_none(self, vectors):
        vectors = vectors.as_lazy()
        col = vectors.ivec[0] > 101
        assert vectors.data[0, 0].compute().shape == (20, 2)
        slic = vectors.ivec[:, col]
        for i in np.ndindex((2, 2)):
            assert slic.data[i].compute().shape == (0, 2)

    def test_row_gte(self, vectors):
        col = vectors.ivec[0] >= 0.5
        assert col.data[0, 0].shape == (20,)
        slic = vectors.ivec[:, col]
        for i in np.ndindex((2, 2)):
            assert np.all(slic.data[i][:, 0] >= 0.5)

    def test_row_lte(self, vectors):
        col = vectors.ivec[0] <= 0.5
        slic = vectors.ivec[:, col]
        for i in np.ndindex((2, 2)):
            assert np.all(slic.data[i][:, 0] <= 0.5)

    def test_vector_slicing_error(self, vectors):
        with pytest.raises(ValueError):
            vectors.ivec[0, 0, 0]

    def test_slicing(self, vectors):
        sliced = vectors.ivec[:, vectors.ivec[0] > 101]
        assert sliced.data[0, 0].shape == (0, 2)

    def test_slicing_first(self, vectors):
        with pytest.raises(ValueError):
            sliced = vectors.ivec[:, 0]

    def test_slicing_lazy(self, vectors):
        vectors = vectors.as_lazy()
        sliced = vectors.ivec[:, vectors.ivec[0] > 101]
        sliced = sliced.ivec[:, sliced.ivec[0] > 101]
        sliced.compute()
        assert sliced.data[0, 0].shape == (0, 2)
        sliced.flatten_diffraction_vectors()

    def test_ragged_vectors(self, vectors):
        vectors = vectors.as_lazy()
        assert vectors.ragged == True

    def test_inav_slicing(self, vectors):
        slic = vectors.inav[0, 0]
        assert isinstance(slic, DiffractionVectors)
        assert slic.data.shape == (1,)

    def test_num_columns(self, vectors):
        assert vectors.num_columns == 2
        lazy_vectors = vectors.as_lazy()
        assert lazy_vectors.num_columns == 2


class TestMask:
    def test_remove_peaks(self):
        s = tilt_boundary_data()
        s.calibration.center = None
        temp = s.template_match_disk(disk_r=5, subtract_min=False)  # Just right
        vectors = temp.get_diffraction_vectors(threshold_abs=0.4, min_distance=5)
        mask = vectors.to_mask(disk_r=12)
        masked_s = s * ~mask
        assert np.max(masked_s.data) < 2



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for operating on insitu signals."""

import numpy as np
import scipy.ndimage as ndi
import scipy.signal as ss


def _register_drift_5d(data, shifts1, shifts2, order=1):
    """
    Register 5D data set using affine transformation

     Parameters
    ----------
    data: np.array or dask.array
        Input image in 5D array (time * rx * ry * kx * ky)
    shifts1: np.array
        1D array for shifts in 1st real space direction or x in hyperspy indexing.
    shifts2: np.array
        1D array for shifts in 2nd real space direction or y in hyperspy indexing.
    order: int
        The order of the spline interpolation for affine transformation. Default
        is 1, has to be in the range 0-5

    Returns
    -------
    data_t: np.array
        5D array after translation according to shift vectors

    """
    data_t = np.zeros_like(data)
    time_size = len(shifts1)
    for i in range(time_size):
        data_t[i, :, :, :, :] = ndi.affine_transform(
            data[i, :, :, :, :],
            np.identity(4),
            offset=(shifts1[i][0, 0, 0, 0], shifts2[i][0, 0, 0, 0], 0, 0),
            order=order,
        )
    return data_t


def _register_drift_2d(data, shift1, shift2, order=1):
    """
    Register 2D data set using affine transformation

     Parameters
    ----------
    data: np.array or dask.array
        Input image in 2D array (ry * rx)
    shift1: float
        shifts in 1st real space direction or x in hyperspy indexing.
    shift2: float
        shifts in 2nd real space direction or y in hyperspy indexing.
    order: int
        The order of the spline interpolation for affine transformation. Default
        is 1, has to be in the range 0-5

    Returns
    -------
    data_t: np.array
        2D array after translation according to shift vectors

    """
    data_t = ndi.affine_transform(
        data, np.identity(2), offset=(shift1, shift2), order=order
    )
    return data_t


def _g2_2d(data, normalization="split", k1bin=1, k2bin=1, tbin=1):
    """
    Calculate k resolved g2(k,t) from I(t,k_r,k_phi)

    Parameters
    ----------
    data: 3D np.array
        Time series for I(t,k_r,k_phi)
    normalization: string
        Normalization format for time autocorrelation, 'split' or 'self'
    k1bin: int
            Binning factor for k1 axis
    k2bin: int
        Binning factor for k2 axis
    tbin: int
        Binning factor for t axis

    Returns
    -------
    g2: 3D np.array
        Time correlation function g2(t,k_r,k_phi)
    """
    data = data.T
    data = (
        data.reshape(
            (data.shape[0] // k1bin),
            k1bin,
            (data.shape[1] // k2bin),
            k2bin,
            (data.shape[2] // tbin),
            tbin,
        )
        .sum(5)
        .sum(3)
        .sum(1)
    )

    # Calculate autocorrelation along time axis
    autocorr = ss.fftconvolve(data, data[:, :, ::-1], mode="full", axes=[-1])
    norm = ss.fftconvolve(np.ones(data.shape), data[:, :, ::-1], mode="full", axes=[-1])
    if normalization == "self":
        overlap_factor = np.expand_dims(
            np.linspace(data.shape[-1], 1, data.shape[-1]), axis=(0, 1)
        )
        norm_factor = norm[:, :, data.shape[-1] : 0 : -1] ** 2
        g2 = autocorr[:, :, data.shape[-1] - 1 :] / norm_factor * overlap_factor
    elif normalization == "split":
        overlap_factor = np.expand_dims(
            np.linspace(data.shape[-1], 1, data.shape[-1]), axis=(0, 1)
        )
        norm_factor = (
            norm[:, :, data.shape[-1] - 1 :]
            * norm[:, :, ::-1][:, :, data.shape[-1] - 1 :]
        )
        g2 = autocorr[:, :, data.shape[-1] - 1 :] / norm_factor * overlap_factor
    else:
        raise ValueError(
            normalization
            + " not recognize, normalization must be chosen 'split' or 'self'"
        )

    return g2.T


def _get_resample_time(t_size, dt, t_rs_size):
    """
    Return log linear resampled time array based on time step and sampling points

    Parameters
    ----------
    t_size: int
        Size of original time array
    dt: float
        Time interval for original time array
    t_rs_size: int
        Size of resampled time array

    Returns
    -------
    t_rs: 1D np.array
        Resampled time array
    """
    t_rs = np.zeros(t_rs_size, dtype=float)

    for i in range(t_rs_size):
        t_rs[i] = np.power(
            10, np.log10(dt) + np.log10(t_size - 1) * i / (t_rs_size - 1)
        )

    return t_rs


def _interpolate_g2_2d(g2, t_rs, dt):
    """
    Interpolate k resolved g2(k,t) based on resampled time

    Parameters
    ----------
    g2: 3D np.array
        Time correlation function g2(t,k_r,k_phi)
    t_rs: 1D np.array
        Resampled time axis array
    dt: float
        Time interval for original g2 function

    Returns
    -------
    g2rs: 3D np.array
        Resampled time correlation function g2(t,k_r,k_phi)
    """
    t = np.round(t_rs / dt, 8)
    g2_l = g2[np.floor(t).astype(int)]
    g2_h = g2[np.ceil(t).astype(int)]
    g2_rs = (
        g2_l + (g2_h - g2_l) * (t - np.floor(t).astype(int))[:, np.newaxis, np.newaxis]
    )
    return g2_rs



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

import pytest

from hyperspy.signals import Signal2D

from pyxem.generators import VirtualDarkFieldGenerator
from pyxem.signals import (
    ElectronDiffraction2D,
    DiffractionVectors,
    VDFSegment,
    DiffractionVectors2D,
    VirtualDarkFieldImage,
)


@pytest.fixture(
    params=[
        np.array(
            [
                np.array([0, 0]),
                np.array([3, 0]),
                np.array([3, 5]),
                np.array([0, 5]),
                np.array([0, 3]),
                np.array([3, 3]),
                np.array([5, 3]),
                np.array([5, 5]),
            ]
        )
    ]
)
def unique_vectors(request):
    uv = DiffractionVectors2D(request.param)
    return uv


@pytest.fixture
def signal_data():
    s = ElectronDiffraction2D(np.zeros((4, 5, 6, 6)))
    s.inav[:2, :2].data[..., 0, 0] = 2
    s.inav[:2, :2].data[..., 0, 3] = 2
    s.inav[:2, :2].data[..., 3, 5] = 2

    s.inav[2:, :3].data[..., 3, 3] = 2
    s.inav[2:, :3].data[..., 3, 0] = 2

    s.inav[2, :2].data[..., 0, 0] = 1
    s.inav[2, :2].data[..., 0, 3] = 1
    s.inav[2, :2].data[..., 3, 5] = 1
    s.inav[2, :2].data[..., 3, 3] = 1
    s.inav[2, :2].data[..., 3, 0] = 1

    s.inav[:2, 2:].data[..., 5, 5] = 3
    s.inav[:2, 2:].data[..., 5, 0] = 3
    s.inav[:2, 2:].data[..., 5, 3] = 3

    s.inav[3:, 2:].data[..., 5, 5] = 3
    s.inav[3:, 2:].data[..., 5, 0] = 3
    return s


@pytest.fixture
def vdf_generator_seg(signal_data, unique_vectors):
    return VirtualDarkFieldGenerator(signal_data, unique_vectors)


@pytest.fixture
def vdf_vector_images_seg(vdf_generator_seg):
    return vdf_generator_seg.get_virtual_dark_field_images(radius=1)


class TestVDFImage:
    def test_init(self):
        vdf = VirtualDarkFieldImage(data=np.ones((8, 10, 10)), vectors=unique_vectors)
        assert isinstance(vdf, VirtualDarkFieldImage)

    @pytest.mark.parametrize(
        "min_distance, min_size, max_size,"
        "max_number_of_grains, marker_radius,"
        "threshold, exclude_border",
        [(1, 1, 20, 5, 1, False, 0), (2, 3, 200, 10, 2, True, 1)],
    )
    def test_get_vdf_segments(
        self,
        vdf_vector_images_seg,
        min_distance,
        min_size,
        max_size,
        max_number_of_grains,
        marker_radius,
        threshold,
        exclude_border,
    ):
        segs = vdf_vector_images_seg.get_vdf_segments(
            min_distance,
            min_size,
            max_size,
            max_number_of_grains,
            marker_radius,
            threshold,
            exclude_border,
        )
        assert isinstance(segs, VDFSegment)
        assert isinstance(segs.segments, Signal2D)
        assert isinstance(segs.vectors_of_segments, DiffractionVectors)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.
import pytest

import numpy as np
from hyperspy._signals.lazy import LazySignal

from pyxem.signals.polar_vectors import PolarVectors, LazyPolarVectors


class TestDiffractionVectors:
    @pytest.fixture
    def polar_vectors(self):
        vectors = np.empty((5,), dtype=object)
        for i in np.ndindex(5):
            phis = np.linspace(0, 2 * np.pi, i[0] + 3)[:-1]  # 2 fold --> 6 fold
            ks = np.ones_like(phis)
            intens = np.ones_like(phis)
            vectors[i] = np.array([ks, phis, intens]).T
        return PolarVectors(
            data=vectors,
            column_names=["k", "phi", "intensity"],
            units=["1/nm", "rad", "a.u."],
        )

    def test_init(self, polar_vectors):
        assert polar_vectors.data.shape == (5,)
        assert polar_vectors.data[0].shape == (2, 3)
        assert polar_vectors.column_names == ["k", "phi", "intensity"]

    def test_get_inscribed_angles_value(self, polar_vectors):
        angles = polar_vectors.get_angles()
        assert angles.data.shape == (5,)
        assert angles.data[0].shape == (0, 5)
        assert angles.data[1].shape == (1, 5)
        for i in range(4):
            np.testing.assert_array_almost_equal(
                angles.data[i + 1][0], np.array([1, 2 * np.pi / (i + 3), 0, 1, 0])
            )

    @pytest.mark.parametrize("intensity_threshold", [None, 0.5])
    @pytest.mark.parametrize("min_k", [None, 0.3])
    def test_get_inscribed_angles(
        self,
        polar_vectors,
        intensity_threshold,
        min_k,
    ):
        polar_vectors.data[2][0, 1] = 2 * np.pi / 3
        angles = polar_vectors.get_angles(
            intensity_threshold=intensity_threshold,
            min_k=min_k,
            accept_threshold=0.1,
            min_angle=0.05,
        )
        assert angles.data.shape == (5,)
        assert angles.data[0].shape == (0, 5)
        assert angles.data[1].shape == (1, 5)
        assert angles.data[2].shape == (1, 5)  # repeated angle

    @pytest.mark.parametrize("cartesian", [True, False])
    def test_get_inscribed_angles_to_markers(self, polar_vectors, cartesian):
        angles = polar_vectors.get_angles()
        angles.to_markers(cartesian=cartesian)

    def test_as_lazy(self, polar_vectors):
        lazy = polar_vectors.as_lazy()
        assert isinstance(lazy, LazySignal)
        assert isinstance(lazy, LazyPolarVectors)
        assert lazy.data.shape == (5,)
        assert lazy.data[0].compute().shape == (2, 3)
        assert lazy.column_names == ["k", "phi", "intensity"]
